from .arrays import Array
from .types import ShapeError

MatchError = ShapeError("The shapes of two vectors do not match")


class Vector(Array):
    """
    It is a vector class that supports basic operations,
    as well as functions such as norm and all angles.
    :param data: An array in the form of a list
    :param check: Check the rationality of the input array
    """

    def __init__(self, data, check=True):
        super().__init__(data, check)
        if len(self.shape) != 1:
            raise ShapeError("Vectors can only be one-dimensional in shape")
        self.len = len(data)

    def __matmul__(self, other):
        if self.len != other.len:
            raise MatchError
        return sum([x * y for x, y in zip(self.data, other.data)])

    def __abs__(self):
        return self.norm()

    def __pos__(self):
        return Vector([x for x in self.data])

    def __neg__(self):
        return Vector([-x for x in self.data])

    def norm(self, p=2):
        if p == 0:
            return len([x for x in self.data if x != 0])
        elif p == 1:
            return sum(map(abs, self.data))
        elif p == 2:
            return sum([x ** 2 for x in self.data]) ** 0.5
        elif p == float("inf"):
            return max(map(abs, self.data))
        else:
            return sum([x ** p for x in self.data]) ** (1 / p)

    def normalize(self):
        norm = self.norm()
        return Vector([x / norm for x in self.data])

    def cosine(self, other):
        if self.len != other.len:
            raise MatchError
        dot_product = self @ other
        norm_self = self.norm()
        norm_other = other.norm()
        return dot_product / (norm_self * norm_other)

    def angles(self, axes=None):
        from math import acos
        identity = []
        for i in range(self.len):
            unit_vector = [0] * self.len
            unit_vector[i] = 1
            identity.append(Vector(unit_vector))
        if axes is None or axes == [] or axes == ():
            result = [acos(self @ axis / (self.norm() * axis.norm())) for axis in identity]
        elif isinstance(axes, int) and axes >= 0:
            axis = identity[axes]
            result = acos(self @ axis / (self.norm() * axis.norm()))
        elif isinstance(axes, (list, tuple)):
            result = [acos(self @ identity[a] / (self.norm() * identity[a].norm())) for a in axes]
        else:
            raise TypeError("The axes parameter can only be None, non-negative integer, list, or tuple")
        return result

    def chebyshev(self, other):
        if self.len != other.len:
            raise MatchError
        return max([abs(x - y) for x, y in zip(self.data, other.data)])

    def manhattan(self, other):
        if self.len != other.len:
            raise MatchError
        return sum([abs(x - y) for x, y in zip(self.data, other.data)])

    def euclidean(self, other):
        if self.len != other.len:
            raise MatchError
        return sum([(x - y) ** 2 for x, y in zip(self.data, other.data)]) ** 0.5

    def minkowski(self, other, p):
        if self.len != other.len:
            raise MatchError
        return sum([abs(x - y) ** p for x, y in zip(self.data, other.data)]) ** (1 / p)


def vec(data):
    return Vector(data)


del Array



from .ufuncs import *


class Group:
    def __init__(self, data, operation=multiply):
        self.__data = set(data)
        self.__operation = None
        self.setop(operation)
        if any([any([not hasattr(elem, attr) for attr in ["__add__", "__sub__", "__mul__", "__truediv__"]])
                for elem in self.__data]):
            raise TypeError("All elements in the group must be computable")

    def __repr__(self):
        return "G{" + ", ".join(sorted([repr(elem) for elem in self.__data])) + "}"

    def elements(self):
        return set(self.__data)

    def getop(self):
        return self.__operation

    def setop(self, operation):
        if not hasattr(operation, "__call__"):
            raise TypeError("The operation function must be callable")
        if operation.__code__.co_argcount != 2:
            raise TypeError("The operation function must have two arguments")
        self.__operation = operation

    def is_closed(self, modulus=None):
        for a in self.__data:
            for b in self.__data:
                result = self.__operation(a, b)
                if modulus is not None:
                    result %= modulus
                if result not in self.__data:
                    return False
        return True

    def is_associative(self, modulus=None):
        for a in self.__data:
            for b in self.__data:
                for c in self.__data:
                    if modulus is None:
                        if self.__operation(self.__operation(a, b), c) != self.__operation(a, self.__operation(b, c)):
                            return False
                    else:
                        if self.__operation(self.__operation(a, b) % modulus, c) % modulus != self.__operation(
                                a, self.__operation(b, c) % modulus) % modulus:
                            return False
        return True

    def has_identity(self):
        identity = None
        for e in self.__data:
            if all([self.__operation(e, a) == a and self.__operation(a, e) == a for a in self.__data]):
                if identity is not None:
                    return False
                identity = e
        return identity is not None

    def has_inverses(self):
        identity = self.identity()
        if identity is None:
            return False
        for a in self.__data:
            for b in self.__data:
                if self.__operation(a, b) == identity and self.__operation(b, a) == identity:
                    break
            else:
                return False
        return True

    def identity(self):
        for e in self.__data:
            if all([self.__operation(e, a) == a and self.__operation(a, e) == a for a in self.__data]):
                return e
        return None

    def is_semigroup(self, modulus=None):
        return True if self.is_closed(modulus) and self.is_associative(modulus) else False

    def is_monoid(self, modulus=None):
        return True if self.is_closed(modulus) and self.is_associative(modulus) and self.has_identity() else False

    def is_group(self, modulus=None):
        return True if all([self.is_closed(modulus), self.is_associative(modulus),
                            self.has_identity(), self.has_inverses()]) else False

    def order(self):
        return len(self.__data)

    def is_supergroup(self, other, modulus=None):
        if not isinstance(other, Group):
            return False
        if not other.is_group(modulus):
            return False
        if self.getop() != other.getop():
            return False
        return self.elements().issuperset(other.elements())

    def is_subgroup(self, other, modulus=None):
        if not isinstance(other, Group):
            return False
        if not other.is_group(modulus):
            return False
        if self.getop() != other.getop():
            return False
        return self.elements().issubset(other.elements())

    def __eq__(self, other):
        return self.__data == other.__data

    def __ne__(self, other):
        return self.__data != other.__data


def group(data):
    return Group(data)



from .types import config, num

SetError = TypeError("The other object must be an OrderedSet")


class OrderedSet:
    def __init__(self, sequence=None):
        """
        It is an ordered set that supports various operations on ordinary sets, as well as other functions.
        :param sequence: A list or tuple
        """
        self.__elements = []
        if sequence is not None:
            self.update(sequence)

    def add(self, element):
        if element not in self.__elements:
            self.__elements.append(element)

    def update(self, elements):
        if isinstance(elements, (list, tuple)):
            for element in elements:
                self.add(element)
        else:
            raise TypeError("Element sequences can only be one list or tuple")

    def insert(self, index, element):
        if element not in self.__elements:
            self.__elements.insert(index, element)

    def pop(self, index=-1):
        try:
            return self.__elements.pop(index)
        except IndexError:
            return None

    def copy(self):
        from copy import deepcopy
        return deepcopy(self)

    def to_list(self):
        from copy import deepcopy
        return list(deepcopy(self.__elements))

    def to_set(self):
        from copy import deepcopy
        return set(deepcopy(self.__elements))

    def hash(self):
        return list(map(hash, self.__elements))

    def find(self, element):
        try:
            return self.__elements.index(element)
        except ValueError:
            return -1

    def __mul__(self, other, replace=False):
        if isinstance(other, OrderedSet):
            if replace:
                self.__elements = [(item1, item2) for item1 in self for item2 in other]
            else:
                result = OrderedSet()
                result.__elements = [(item1, item2) for item1 in self for item2 in other]
                return result
        else:
            raise SetError

    cartesian_product = __mul__

    def __imul__(self, other):
        self.__mul__(other, True)
        return self

    def powerset(self):
        n = self.__len__()
        power_set = []
        for i in range(1 << n):
            subset = []
            for j in range(n):
                if i & 1 << j:
                    subset.append(self[j])
            result = OrderedSet()
            result.__elements = subset
            power_set.append(result)
        result = OrderedSet()
        result.__elements = power_set
        return result

    def partitions(self, k=None):
        def backtrack(start, path, res):
            if start == len(self):
                if not k or len(path) == k:
                    res.append(path)
                return
            backtrack(start + 1, path + [[self[start]]], res)
            for i, subset in enumerate(path):
                new_subset = subset + [self[start]]
                path_copy = path[:]
                path_copy[i] = new_subset
                backtrack(start + 1, path_copy, res)

        if not self.__elements:
            return [[]] if k is None or k == 0 else []
        result = []
        backtrack(0, [[]], result)
        final_result = []
        for p in result:
            if k is None or len(p) == k:
                temp_list = []
                for s in p:
                    temp = OrderedSet()
                    temp.__elements = s
                    temp_list.append(temp)
                final_result.append(temp_list)
        return sorted(final_result) if k is None else final_result

    def sort(self, key=None, reverse=False, new=True):
        try:
            if new:
                result = OrderedSet()
                result.__elements = sorted(self.__elements, key=key, reverse=reverse)
                return result
            self.__elements = sorted(self.__elements, key=key, reverse=reverse)
        except TypeError:
            return self

    def remove(self, element):
        self.__elements.remove(element)

    def clear(self):
        self.__elements = []

    def __contains__(self, element):
        return element in self.__elements

    def __getitem__(self, item):
        return self.__elements[item]

    def __len__(self):
        return len(self.__elements)

    def __sub__(self, other, replace=False):
        if isinstance(other, OrderedSet):
            result = list(filter(lambda e: e not in other.__elements, self.__elements))
            if replace:
                self.__elements = result
            else:
                return OrderedSet(result)
        else:
            raise SetError

    def __and__(self, other, replace=False):
        if isinstance(other, OrderedSet):
            total = self.__elements + other.__elements
            filtered = list(filter(lambda e: total.count(e) != 1, total))
            result = filtered[:len(filtered) // 2]
            if replace:
                self.__elements = result
            else:
                return OrderedSet(result)
        else:
            raise SetError

    def __or__(self, other, replace=False):
        if isinstance(other, OrderedSet):
            if replace:
                self.update(other.__elements)
            else:
                return OrderedSet(self.__elements + other.__elements)
        else:
            raise SetError

    def __matmul__(self, other, replace=False):
        if isinstance(other, OrderedSet):
            result = list(filter(lambda e: e not in self.__elements, other.__elements))
            if replace:
                self.__elements = result
            else:
                return OrderedSet(result)
        else:
            raise SetError

    def __xor__(self, other, replace=False):
        if isinstance(other, OrderedSet):
            total = self.__elements + other.__elements
            result = list(filter(lambda e: total.count(e) != 2, total))
            if replace:
                self.__elements = result
            else:
                return OrderedSet(result)
        else:
            raise SetError

    difference = __sub__
    intersection = __and__
    union = __or__
    complement = __matmul__
    symmetric_difference = __xor__

    def __isub__(self, other):
        self.__sub__(other, True)
        return self

    def __iand__(self, other):
        self.__and__(other, True)
        return self

    def __ior__(self, other):
        self.__or__(other, True)
        return self

    def __imatmul__(self, other):
        self.__matmul__(other, True)
        return self

    def __ixor__(self, other):
        self.__xor__(other, True)
        return self

    def __gt__(self, other):
        if not isinstance(other, OrderedSet):
            raise SetError
        return self.to_set() > other.to_set()

    def __ge__(self, other):
        if not isinstance(other, OrderedSet):
            raise SetError
        return self.to_set() >= other.to_set()

    def __eq__(self, other):
        if not isinstance(other, OrderedSet):
            raise SetError
        return self.to_set() == other.to_set()

    def __le__(self, other):
        if not isinstance(other, OrderedSet):
            raise SetError
        return self.to_set() <= other.to_set()

    def __lt__(self, other):
        if not isinstance(other, OrderedSet):
            raise SetError
        return self.to_set() < other.to_set()

    def __ne__(self, other):
        if not isinstance(other, OrderedSet):
            raise SetError
        return self.to_set() != other.to_set()

    def is_proper_superset(self, other):
        return self > other

    def is_superset(self, other):
        return self >= other

    def is_equal(self, other):
        return self == other

    def is_subset(self, other):
        return self <= other

    def is_proper_subset(self, other):
        return self < other

    def is_not_equal(self, other):
        return self != other

    def __hash__(self):
        return hash(self.__str__())

    def __repr__(self):
        return "OrderedSet({" + str(self.__elements)[1:-1] + "})"


class InfIterator:
    arithmetic = "arithmetic"
    geometric = "geometric"
    fibonacci = "fibonacci"
    catalan = "catalan"

    def __init__(self, start: num = 0, mode: str = "", common: num = 2):
        """
        Infinite sequence iterator, default to self increasing sequence.
        :param start: Initial value
        :param mode: Iterative mode
        :param common: Tolerance or common ratio
        """
        self.mode = mode
        self.common = common
        if self.mode == self.fibonacci:
            self.current = 0
        elif self.mode == self.catalan:
            self.current = 1
        else:
            self.current = start
        self.__next_value = 1 if self.mode == self.fibonacci or self.mode == self.catalan else 0

    def __iter__(self):
        return self

    def __next__(self):
        value = self.current
        if self.mode == self.arithmetic:
            self.current += self.common
        elif self.mode == self.geometric:
            self.current *= self.common
        elif self.mode == self.fibonacci:
            self.__next_value, self.current = self.current + self.__next_value, self.__next_value
        elif self.mode == self.catalan:
            self.current = self.current * (4 * self.__next_value - 2) // (self.__next_value + 1)
            self.__next_value += 1
        else:
            self.current += 1
        return value

    def __repr__(self):
        return "InfIterator(" + self.mode + ")"


class LinkedListNode:
    def __init__(self, value, next_node=None):
        """
        The nodes of a linked list can be connected by setting up successor nodes.
        :param value: The value of the node
        :param next_node: Next node object
        """
        self.value = value
        self.next_node = next_node


class LinkedList:
    def __init__(self):
        """
        It is a linked list that supports many functions.
        """
        self.head = None

    def length(self):
        result = 0
        current = self.head
        while current is not None:
            result += 1
            current = current.next_node
        return result

    def insert_beg(self, value):
        new_node = LinkedListNode(value)
        new_node.next_node = self.head
        self.head = new_node

    def insert_end(self, value):
        if self.head is None:
            self.head = LinkedListNode(value)
        else:
            current = self.head
            while current.next_node is not None:
                current = current.next_node
            current.next_node = LinkedListNode(value)

    def insert_idx(self, index, value):
        if index < 0:
            raise IndexError("Index must be non-negative")
        if index == 0:
            self.insert_beg(value)
            return
        current = self.head
        position = 0
        while current is not None and position < index - 1:
            current = current.next_node
            position += 1
        if not current:
            raise IndexError("Index out of range")
        new_node = LinkedListNode(value)
        new_node.next_node = current.next_node
        current.next_node = new_node

    def insert_after(self, target_value, new_value):
        if self.head is None:
            raise ValueError("The linked list is empty")
        current = self.head
        found = False
        while current is not None and not found:
            if current.value == target_value:
                found = True
            else:
                current = current.next_node
        if not found:
            raise ValueError("No node with the specified value was found")
        new_node = LinkedListNode(new_value)
        new_node.next_node = current.next_node
        current.next_node = new_node

    def replace(self, target_value, new_value, n=1):
        current = self.head
        count = 0
        while current is not None and count < n:
            if current.value == target_value:
                current.value = new_value
                count += 1
            current = current.next_node

    def delete_node(self, value):
        if self.head is None:
            return
        if self.head.value == value:
            self.head = self.head.next_node
            return
        current = self.head
        while current.next_node is not None:
            if current.next_node.value == value:
                current.next_node = current.next_node.next_node
                return
            current = current.next_node

    def contains(self, value):
        current = self.head
        while current is not None:
            if current.value == value:
                return True
            current = current.next_node
        return False

    def count(self, value):
        result = 0
        current = self.head
        while current is not None:
            if current.value == value:
                result += 1
            current = current.next_node
        return result

    def get_successor(self, value):
        current = self.head
        while current is not None and current.value != value:
            current = current.next_node
        if current is not None and current.next_node is not None:
            return current.next_node.value

    def to_list(self):
        result = []
        current = self.head
        while current:
            result.append(current.value)
            current = current.next_node
        return result

    def __len__(self):
        return self.length()

    def __contains__(self, item):
        return self.contains(item)

    def __repr__(self):
        current = self.head
        linked_list_str = ""
        while current is not None:
            linked_list_str += str(current.value)
            if current.next_node:
                linked_list_str += " -> "
            current = current.next_node
        return linked_list_str


class IntervalSet:
    def __init__(self, intervals=None):
        """
        It is a set of intervals and supports multiple operations.
        :param intervals: A sequence containing multiple intervals
        """
        self.intervals = []
        if intervals:
            self.add_intervals(intervals)

    def add_intervals(self, intervals):
        for interval in sorted(intervals, key=lambda x: x[0]):
            self.add_interval(*interval)

    def add_interval(self, start, end, start_open=False, end_open=False):
        if start == float("-inf") or start == float("inf"):
            start_open = True
        if end == float("-inf") or end == float("inf"):
            end_open = True
        start_open = bool(start_open)
        end_open = bool(end_open)
        if start > end or (start == end and (start_open or end_open)) or start == end == float(
                "inf") or start == end == float("-inf"):
            return
        intervals = []
        added = False
        for s, e, so, eo in self.intervals:
            if s <= start < e or s < end <= e or start <= s < end or start < e <= end or start == e and not (
                    start_open and eo) or s == end and not (so and end_open):
                start = min(s, start)
                end = max(e, end)
                start_open = start_open if start < s else so if start > s else start_open and so
                end_open = end_open if end > e else eo if end < e else end_open and eo
            elif start > e:
                intervals.append((s, e, so, eo))
            else:
                if not added:
                    intervals.append((start, end, start_open, end_open))
                    added = True
                intervals.append((s, e, so, eo))
        if not added:
            intervals.append((start, end, start_open, end_open))
        self.intervals = intervals

    def remove_interval(self, start, end, start_open=False, end_open=False):
        intervals = []
        start_open = bool(start_open)
        end_open = bool(end_open)
        for s, e, so, eo in self.intervals:
            if e <= start:
                intervals.append((s, e, so, eo))
            elif s >= end:
                intervals.append((s, e, so, eo))
            else:
                if s < start:
                    intervals.append((s, start, so, not start_open))
                if e > end:
                    intervals.append((end, e, not end_open, eo))
        self.intervals = intervals

    def is_superset(self, other):
        def contains_interval(a, b):
            s1, e1, so1, eo1 = a
            s2, e2, so2, eo2 = b
            return (s2 <= s1 if not so2 else s2 < s1) and (e1 <= e2 if not eo1 else e1 < e2)

        for interval in other.intervals:
            if not any([contains_interval(interval, other_interval) for other_interval in self.intervals]):
                return False
        return True

    def is_proper_superset(self, other):
        return self.is_superset(other) and self.is_not_equal(other)

    def is_equal(self, other):
        return self.intervals == other.intervals

    def is_proper_subset(self, other):
        return other.is_proper_superset(self)

    def is_subset(self, other):
        return other.is_superset(self)

    def is_not_equal(self, other):
        return self.intervals != other.intervals

    __gt__ = is_proper_superset
    __ge__ = is_superset
    __eq__ = is_equal
    __le__ = is_subset
    __lt__ = is_proper_subset
    __ne__ = is_not_equal

    def intersection(self, other):
        result = IntervalSet()
        i, j = 0, 0
        while i < len(self.intervals) and j < len(other.intervals):
            s1, e1, so1, eo1 = self.intervals[i]
            s2, e2, so2, eo2 = other.intervals[j]
            start = max(s1, s2)
            end = min(e1, e2)
            start_open = so1 if s1 > s2 else so2 if s1 < s2 else so1 or so2
            end_open = eo1 if e1 < e2 else eo2 if e1 > e2 else eo1 or eo2
            if start < end or (start == end and not (start_open and end_open)):
                result.add_interval(start, end, start_open, end_open)
            if e1 <= e2:
                i += 1
            if e2 <= e1:
                j += 1
        return result

    def union(self, other):
        result = IntervalSet(self.intervals)
        for interval in other.intervals:
            result.add_interval(*interval)
        return result

    def difference(self, other):
        result = IntervalSet(self.intervals)
        for s, e, so, eo in other.intervals:
            result.remove_interval(s, e, so, eo)
        return result

    def complement(self, other):
        return other.difference(self)

    def symmetric_difference(self, other):
        return self.union(other).difference(self.intersection(other))

    __and__ = intersection
    __or__ = union
    __sub__ = difference
    __matmul__ = complement
    __xor__ = symmetric_difference

    def latex(self):
        return self.__repr__(True)

    def __repr__(self, use_latex=False):
        use_latex = config.use_latex or use_latex
        intervals = self.intervals
        n = len(intervals)
        if n == 0:
            if use_latex:
                return "\\emptyset"
            elif config.use_unicode:
                return "∅"
            else:
                return "EmptySet"

        def format_interval(s, e, so, eo):
            infinity = "\\infty" if use_latex else "∞" if config.use_unicode else "inf"
            if s == e:
                if use_latex:
                    return "\\left\\{" + str(s) + "\\right\\}"
                else:
                    return "{{{}}}".format(s)
            else:
                start_str = str(s).replace("inf", infinity)
                end_str = str(e).replace("inf", infinity)
                if use_latex:
                    start_bracket = "\\left(" if so else "\\left["
                    end_bracket = "\\right)" if eo else "\\right]"
                else:
                    start_bracket = "(" if so else "["
                    end_bracket = ")" if eo else "]"
                return "{}{}, {}{}".format(start_bracket, start_str, end_str, end_bracket)

        formatted_intervals = [format_interval(*interval) for interval in intervals]
        union = " \\cup " if use_latex else " ∪ " if config.use_unicode else " U "
        return union.join(formatted_intervals)



"""
This module defines a collection of constants representing various physical, mathematical, and unit conversion
factors. These constants are commonly used in scientific and engineering calculations.

- **EB**: Exabyte, equal to 2^60 bytes.
- **G**: Gravitational constant (m^3 kg^-1 s^-2).
- **GB**: Gigabyte, equal to 2^30 bytes.
- **KB**: Kilobyte, equal to 2^10 bytes.
- **MB**: Megabyte, equal to 2^20 bytes.
- **PB**: Petabyte, equal to 2^50 bytes.
- **TB**: Terabyte, equal to 2^40 bytes.
- **YB**: Yottabyte, equal to 2^80 bytes.
- **ZB**: Zettabyte, equal to 2^70 bytes.
- **a_0**: Bohr radius (meters).
- **acre**: Area of one acre (square meters).
- **alpha**: Fine-structure constant.
- **alpha_f**: Feigenbaum constant.
- **amu**: Atomic mass unit (kilograms).
- **apery**: Apery's constant.
- **arcmin**: Minute of arc (degrees).
- **arcsec**: Second of arc (degrees).
- **atm**: Standard atmosphere (pascals).
- **atto**: Prefix denoting 10^-18.
- **au**: Astronomical unit (meters).
- **btu**: British thermal unit (joules).
- **c**: Speed of light in a vacuum (meters per second).
- **cal**: Calorie (joules).
- **catalan**: Catalan's constant.
- **centi**: Prefix denoting 10^-2.
- **ci**: Curie (becquerels).
- **conway**: Conway's constant.
- **ct**: Carat (kilograms).
- **cup**: Volume of one US cup (cubic meters).
- **day**: Number of seconds in one day.
- **debye**: Unit of electric dipole moment (coulomb-meters).
- **deci**: Prefix denoting 10^-1.
- **deka**: Prefix denoting 10^1.
- **delta_f**: Feigenbaum constant delta.
- **e**: Euler's number.
- **epsilon_0**: Vacuum permittivity (farads per meter).
- **exa**: Prefix denoting 10^18.
- **faraday**: Faraday constant (coulombs).
- **fathom**: Length of one fathom (meters).
- **fc**: Foot-candle (lumens per square meter).
- **femto**: Prefix denoting 10^-15.
- **fl**: Foot-lambert (candelas per square meter).
- **fl_oz**: Fluid ounce (cubic meters).
- **foias**: Foias constant.
- **foot**: Length of one foot (meters).
- **g**: Standard acceleration due to gravity (m/s^2).
- **gal_can**: Canadian gallon (cubic meters).
- **gal_uk**: UK gallon (cubic meters).
- **gal_us**: US gallon (cubic meters).
- **gamma**: Euler-Mascheroni constant.
- **gauss**: Unit of magnetic field strength (tesla).
- **giga**: Prefix denoting 10^9.
- **glaisher**: Glaisher-Kinkelin constant.
- **h**: Planck constant (joule-seconds).
- **hbar**: Reduced Planck constant (joule-seconds).
- **hecto**: Prefix denoting 10^2.
- **hour**: Number of seconds in one hour.
- **hp**: Horsepower (watts).
- **inch**: Length of one inch (meters).
- **inf**: Infinity (float representation).
- **inh2o**: Inches of water (pascals).
- **inhg**: Inches of mercury (pascals).
- **k_b**: Boltzmann constant (joules per kelvin).
- **k_e**: Coulomb's constant (newton-meter^2 per coulomb^2).
- **kilo**: Prefix denoting 10^3.
- **kip**: Kip (force) (newtons).
- **kmh**: Kilometers per hour (meters per second).
- **knot**: Nautical mile per hour (meters per second).
- **lbf**: Pound-force (newtons).
- **lbm**: Pound-mass (kilograms).
- **ly**: Light-year (meters).
- **m_e**: Electron mass (kilograms).
- **m_mu**: Muon mass (kilograms).
- **m_n**: Neutron mass (kilograms).
- **m_p**: Proton mass (kilograms).
- **m_sun**: Solar mass (kilograms).
- **mega**: Prefix denoting 10^6.
- **micro**: Prefix denoting 10^-6.
- **mil**: Thousandth of an inch (meters).
- **mile**: Length of one mile (meters).
- **milli**: Prefix denoting 10^-3.
- **minute**: Number of seconds in one minute.
- **mmhg**: Millimeters of mercury (pascals).
- **mph**: Miles per hour (meters per second).
- **mu_0**: Vacuum permeability (henry per meter).
- **mu_b**: Bohr magneton (joules per tesla).
- **mu_e**: Electron magnetic moment (joules per tesla).
- **mu_n**: Neutron magnetic moment (joules per tesla).
- **mu_p**: Proton magnetic moment (joules per tesla).
- **n_a**: Avogadro constant (per mole).
- **nan**: Not a number (float representation).
- **nano**: Prefix denoting 10^-9.
- **nmi**: Nautical mile (meters).
- **oz_t**: Troy ounce (kilograms).
- **ozm**: Ounce-mass (kilograms).
- **parsec**: Parsec (meters).
- **pdl**: Poundal (force) (newtons).
- **peta**: Prefix denoting 10^15.
- **phi**: Golden ratio.
- **pi**: Pi, the ratio of a circle's circumference to its diameter.
- **pico**: Prefix denoting 10^-12.
- **pint**: Volume of one pint (cubic meters).
- **point**: Point (typography) (meters).
- **psi**: Pounds per square inch (pascals).
- **q_e**: Elementary charge (coulombs).
- **quart**: Volume of one quart (cubic meters).
- **r_g**: Universal gas constant (joules per mole kelvin).
- **radian**: Radians in one degree.
- **roentgen**: Roentgen (exposure to ionizing radiation).
- **ry**: Rydberg constant (energy).
- **sigma**: Stefan-Boltzmann constant (watts per square meter per kelvin^4).
- **sigma_t**: Thomson cross-section (square meters).
- **tbsp**: Tablespoon (cubic meters).
- **tera**: Prefix denoting 10^12.
- **texpoint**: Tex point (typography) (meters).
- **therm**: Therm (energy).
- **torr**: Torr (pressure).
- **tsp**: Teaspoon (cubic meters).
- **uk_ton**: UK ton (kilograms).
- **us_ton**: US ton (kilograms).
- **v_m**: Molar volume (cubic meters per mole).
- **week**: Number of seconds in one week.
- **yard**: Length of one yard (meters).
- **year**: Number of seconds in one year.
- **yocto**: Prefix denoting 10^-24.
- **yotta**: Prefix denoting 10^24.
- **zepto**: Prefix denoting 10^-21.
- **zetta**: Prefix denoting 10^21.
"""

EB = 1152921504606846976
G = 6.6743e-11
GB = 1073741824
KB = 1024
MB = 1048576
PB = 1125899906842624
TB = 1099511627776
YB = 1208925819614629174706176
ZB = 1180591620717411303424
a_0 = 5.291772083e-11
acre = 4046.85642241
alpha = 0.0072973525693
alpha_f = 2.5029078750958926
amu = 1.6605402e-27
apery = 1.2020569031595942
arcmin = 0.016666666666666666
arcsec = 0.0002777777777777778
atm = 101325
atto = 1e-18
au = 149597870700
btu = 1055.05585262
c = 299792458
cal = 4.1868
catalan = 0.915965594177219
centi = 0.01
ci = 37000000000
conway = 1.3035772690342964
ct = 0.0002
cup = 0.000236588236501
day = 86400
debye = 3.33564095198e-30
deci = 0.1
deka = 10
delta_f = 4.66920160910299
e = 2.718281828459045
epsilon_0 = 8.854187817620389e-12
exa = 1000000000000000000
faraday = 96485.3429775
fathom = 1.8288
fc = 10.76
femto = 1e-15
fl = 10.7639104
fl_oz = 2.95735295626e-05
foias = 1.1874523510652712
foot = 0.3048
g = 9.80665
gal_can = 0.00454609
gal_uk = 0.004546092
gal_us = 0.00378541178402
gamma = 0.5772156649015329
gauss = 0.8346268416740732
giga = 1000000000
glaisher = 1.2824271291006226
h = 6.62607015e-34
hbar = 1.0545718176461565e-34
hecto = 100
hour = 3600
hp = 745.6998715822701
inch = 0.0254
inf = float("inf")
inh2o = 249.0889
inhg = 3386.38815789
k_b = 1.380649e-23
k_e = 8.9875517923e+9
kilo = 1000
kip = 4448.22161526
kmh = 0.2777777777777778
knot = 0.5144444444444445
lbf = 4.44822161526
lbm = 0.45359237
ly = 9460730472580800
m_e = 9.1093837015e-31
m_mu = 1.88353109e-28
m_n = 1.67492749804e-27
m_p = 1.67262192369e-27
m_sun = 1.98892e+30
mega = 1000000
micro = 1e-06
mil = 2.54e-05
mile = 1609.344
milli = 0.001
minute = 60
mmhg = 133322.368421
mph = 0.44704
mu_0 = 1.2566370614359173e-06
mu_b = 9.27400899e-24
mu_e = 9.28476362e-24
mu_n = 5.05078317e-27
mu_p = 1.410606633e-26
n_a = 6.02214076e+23
nan = float("nan")
nano = 1e-09
nmi = 1852
oz_t = 0.031103475
ozm = 0.028349523125
parsec = 3.085677581491367e+16
pdl = 0.138255
peta = 1000000000000000
phi = 1.618033988749895
pi = 3.141592653589793
pico = 1e-12
pint = 0.000473176473002
point = 0.00035277777777777776
psi = 6894.75729317
q_e = 1.602176634e-19
quart = 0.000946352946004
r_g = 8.314462618
radian = 57.29577951308232
roentgen = 0.000258
ry = 2.17987196968e-18
sigma = 5.67040047374e-08
sigma_t = 6.65245893699e-29
tbsp = 1.47867647813e-05
tera = 1000000000000
texpoint = 0.000351459803515
therm = 105506000
torr = 133.322368421
tsp = 4.92892159375e-06
uk_ton = 1016.0469088
us_ton = 907.18474
v_m = 0.022710981
week = 604800
yard = 0.9144
year = 31557600
yocto = 1e-24
yotta = 1000000000000000000000000
zepto = 1e-21
zetta = 1000000000000000000000



from .types import LogicError

FullError = LogicError("The input of this logic component is full")
InputError = LogicError("The output port number of the connection must be an integer")


class Basic:
    def __init__(self, label):
        if not isinstance(label, str):
            raise TypeError("The type of label can only be a string")
        self.type = str(type(self))
        self.__label = label
        self.data = None
        self.order0 = 0
        self.order1 = 0
        self.order2 = 0
        self.order3 = 0

    def __repr__(self):
        return self.__class__.__name__ + "({})".format(self.__label)

    def out(self):
        raise NotImplementedError

    def set_order0(self, data):
        if not isinstance(data, int):
            raise InputError
        self.order0 = data

    def set_order1(self, data):
        if isinstance(self, Unary):
            raise NotImplementedError
        if not isinstance(data, int):
            raise InputError
        self.order1 = data

    def set_order2(self, data):
        if isinstance(self, Unary) or isinstance(self, Binary):
            raise NotImplementedError
        if not isinstance(data, int):
            raise InputError
        self.order2 = data

    def set_order3(self, data):
        if isinstance(self, Unary) or isinstance(self, Binary) or isinstance(self, Ternary):
            raise NotImplementedError
        if not isinstance(data, int):
            raise InputError
        self.order3 = data


class Unary(Basic):
    def __init__(self, label, pin0=None):
        super().__init__(label)
        if pin0 is not None and not isinstance(pin0, int):
            raise InputError
        self.pin0 = pin0

    def set_pin0(self, data):
        if not isinstance(data, int) and not isinstance(data, Basic):
            raise InputError
        self.pin0 = data

    def out(self):
        self.data = self.pin0.out()[
            self.order0] if isinstance(self.pin0, Basic) else 0 if self.pin0 is None else self.pin0


class Binary(Basic):
    def __init__(self, label, pin0=None, pin1=None):
        super().__init__(label)
        if any(map(lambda item: item is not None and not isinstance(item, int), [pin0, pin1])):
            raise InputError
        self.pin0 = pin0
        self.pin1 = pin1

    def set_pin0(self, data):
        if not isinstance(data, int) and not isinstance(data, Basic):
            raise InputError
        self.pin0 = data

    def set_pin1(self, data):
        if not isinstance(data, int) and not isinstance(data, Basic):
            raise InputError
        self.pin1 = data

    def out(self):
        self.data = [
            self.pin0.out()[self.order0] if isinstance(self.pin0, Basic) else 0 if self.pin0 is None else self.pin0,
            self.pin1.out()[self.order1] if isinstance(self.pin1, Basic) else 0 if self.pin1 is None else self.pin1
        ]


class Ternary(Basic):
    def __init__(self, label, pin0=None, pin1=None, pin2=None):
        super().__init__(label)
        if any(map(lambda item: item is not None and not isinstance(item, int), [pin0, pin1, pin2])):
            raise InputError
        self.pin0 = pin0
        self.pin1 = pin1
        self.pin2 = pin2

    def set_pin0(self, data):
        if not isinstance(data, int) and not isinstance(data, Basic):
            raise InputError
        self.pin0 = data

    def set_pin1(self, data):
        if not isinstance(data, int) and not isinstance(data, Basic):
            raise InputError
        self.pin1 = data

    def set_pin2(self, data):
        if not isinstance(data, int) and not isinstance(data, Basic):
            raise InputError
        self.pin2 = data

    def out(self):
        self.data = [
            self.pin0.out()[self.order0] if isinstance(self.pin0, Basic) else 0 if self.pin0 is None else self.pin0,
            self.pin1.out()[self.order1] if isinstance(self.pin1, Basic) else 0 if self.pin1 is None else self.pin1,
            self.pin2.out()[self.order2] if isinstance(self.pin2, Basic) else 0 if self.pin2 is None else self.pin2
        ]


class Quaternary(Basic):
    def __init__(self, label, pin0=None, pin1=None, pin2=None, pin3=None):
        super().__init__(label)
        if any(map(lambda item: item is not None and not isinstance(item, int), [pin0, pin1, pin2, pin3])):
            raise InputError
        self.pin0 = pin0
        self.pin1 = pin1
        self.pin2 = pin2
        self.pin3 = pin3

    def set_pin0(self, data):
        if not isinstance(data, int) and not isinstance(data, Basic):
            raise InputError
        self.pin0 = data

    def set_pin1(self, data):
        if not isinstance(data, int) and not isinstance(data, Basic):
            raise InputError
        self.pin1 = data

    def set_pin2(self, data):
        if not isinstance(data, int) and not isinstance(data, Basic):
            raise InputError
        self.pin2 = data

    def set_pin3(self, data):
        if not isinstance(data, int) and not isinstance(data, Basic):
            raise InputError
        self.pin3 = data

    def out(self):
        self.data = [
            self.pin0.out()[self.order0] if isinstance(self.pin0, Basic) else 0 if self.pin0 is None else self.pin0,
            self.pin1.out()[self.order1] if isinstance(self.pin1, Basic) else 0 if self.pin1 is None else self.pin1,
            self.pin2.out()[self.order2] if isinstance(self.pin2, Basic) else 0 if self.pin2 is None else self.pin2,
            self.pin3.out()[self.order3] if isinstance(self.pin3, Basic) else 0 if self.pin3 is None else self.pin3
        ]


def connector(previous, latter):
    if not isinstance(previous, Basic) or not isinstance(latter, Basic):
        raise TypeError("The connected logical components must inherit from the Basic class")
    previous = previous
    latter = latter
    if isinstance(latter, Unary):
        if latter.pin0 is None:
            latter.set_pin0(previous)
        else:
            raise FullError
    elif isinstance(latter, Binary):
        if latter.pin0 is None:
            latter.set_pin0(previous)
        elif latter.pin1 is None:
            latter.set_pin1(previous)
        else:
            raise FullError
    elif isinstance(latter, Ternary):
        if latter.pin0 is None:
            latter.set_pin0(previous)
        elif latter.pin1 is None:
            latter.set_pin1(previous)
        elif latter.pin2 is None:
            latter.set_pin2(previous)
        else:
            raise FullError
    else:
        raise NotImplementedError


class NOT(Unary):
    def out(self):
        super().out()
        return [1 - self.data]


class AND(Binary):
    def out(self):
        super().out()
        return [1 if all(self.data) else 0]


class OR(Binary):
    def out(self):
        super().out()
        return [1 if any(self.data) else 0]


class NAND(Binary):
    def out(self):
        super().out()
        return [0 if all(self.data) else 1]


class NOR(Binary):
    def out(self):
        super().out()
        return [0 if any(self.data) else 1]


class XNOR(Binary):
    def out(self):
        super().out()
        return [1 if self.data[0] == self.data[1] else 0]


class XOR(Binary):
    def out(self):
        super().out()
        return [0 if self.data[0] == self.data[1] else 1]


class HalfAdder(Binary):
    def out(self):
        super().out()
        return [0 if self.data[0] == self.data[1] else 1, 1 if all(self.data) else 0]


class FullAdder(Ternary):
    def out(self):
        super().out()
        return [0 if sum(self.data) % 2 == 0 else 1, 1 if sum(self.data) > 1 else 0]


class HalfSuber(Binary):
    def out(self):
        super().out()
        return [0 if self.data[0] == self.data[1] else 1, 1 if not self.data[0] and self.data[1] else 0]


class FullSuber(Ternary):
    def out(self):
        super().out()
        return [0 if sum(self.data) % 2 == 0 else 1, 1 if self.data[0] - self.data[1] - self.data[2] < 0 else 0]


class TwoBMuler(Quaternary):
    def out(self):
        super().out()
        if self.data[0] == 0 and self.data[1] == 0 or self.data[2] == 0 and self.data[3] == 0:
            return [0, 0, 0, 0]
        elif self.data[0] == 1 and self.data[1] == 0:
            return [self.data[2], self.data[3], 0, 0]
        elif self.data[2] == 1 and self.data[3] == 0:
            return [self.data[0], self.data[1], 0, 0]
        elif self.data[0] == 0 and self.data[1] == 1:
            return [0, self.data[2], self.data[3], 0]
        elif self.data[2] == 0 and self.data[3] == 1:
            return [0, self.data[0], self.data[1], 0]
        else:
            return [1, 0, 0, 1]


class TwoBDiver(Quaternary):
    def out(self):
        super().out()
        if self.data[2] == 0 and self.data[3] == 0:
            return [1, 1, 1, 1]
        elif self.data[0] == 0 and self.data[1] == 0:
            return [0, 0, 0, 0]
        elif self.data[2] == 1 and self.data[3] == 0:
            return [self.data[0], self.data[1], 0, 0]
        elif self.data[2] == 0 and self.data[3] == 1:
            return [self.data[1], 0, self.data[0], 0]
        elif self.data == [1, 1, 1, 1]:
            return [1, 0, 0, 0]
        else:
            return [0, 0, self.data[0], self.data[1]]


class JKFF(Binary):
    def __init__(self, label, pin0=None, pin1=None, state=0):
        super().__init__(label, pin0, pin1)
        self.__state = state

    def out(self):
        super().out()
        j, k = self.data
        if j == 0 and k == 0:
            self.__state = self.__state
        elif j == 0 and k == 1:
            self.__state = 0
        elif j == 1 and k == 0:
            self.__state = 1
        elif j == 1 and k == 1:
            self.__state = 1 - self.__state
        return [self.__state]


class DFF(Unary):
    def __init__(self, label, pin0=None, state=0):
        super().__init__(label, pin0)
        self.__state = state

    def out(self):
        super().out()
        d = self.data
        if d == 0:
            self.__state = 0
        elif d == 1:
            self.__state = 1
        return [self.__state]


class TFF(Unary):
    def __init__(self, label, pin0=None, state=0):
        super().__init__(label, pin0)
        self.__state = state

    def out(self):
        super().out()
        t = self.data
        if t == 0:
            self.__state = self.__state
        elif t == 1:
            self.__state = 1 - self.__state
        return [self.__state]


class COMP(Binary):
    def out(self):
        super().out()
        return [1, 0, 0] if self.data[0] and not self.data[1] else [
            0, 0, 1] if not self.data[0] and self.data[1] else [0, 1, 0]



from .types import Any, Callable, ShapeError, arr

ArrayError = ShapeError("The shape of the array is invalid")
MatchError = ShapeError("The shapes of the two arrays do not match")


class Array:
    """
    It is the base class of vectors and matrices, supporting operations and many statistical functions.
    :param data: An array in the form of a list
    :param check: Check the rationality of the input array
    """

    def __init__(self, data=None, check=True):
        if data is None:
            data = []
        self.shape = [] if data == [] else get_shape(data)
        if check and self.shape and not isinstance(data, (int, float, complex)):
            is_valid_array(data, self.shape)
        self.data = data

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.data)

    def __str__(self):
        if not self.data:
            return "[]"

        def _format(_nested_list, _max_length):
            if isinstance(_nested_list, list):
                _copy = []
                for item in _nested_list:
                    _copy.append(_format(item, _max_length))
                return _copy
            else:
                _item = repr(_nested_list)
                return " " * (_max_length - len(_item)) + _item

        _max = len(max(str(self.data).replace("[", "").replace("]", "").replace(",", "").split(), key=len))
        _then = str(_format(self.data, _max)).replace("], ", "]\n").replace(",", "").replace("'", "").split("\n")
        _max = max([_.count("[") for _ in _then])
        return "\n".join([(_max - _.count("[")) * " " + _ + "\n" * (_.count("]") - 1) for _ in _then]).strip()

    def __getitem__(self, item):
        def get_item_recursive(result, indices):
            if indices == ():
                return result
            index = indices[0]
            if isinstance(index, int):
                return get_item_recursive(result[index], indices[1:])
            elif isinstance(index, (slice, list, tuple, range)):
                if isinstance(index, slice):
                    start = index.start
                    stop = index.stop
                    step = index.step
                    if step is None or step > 0:
                        start = 0 if start is None else start
                        stop = len(result) if stop is None else stop
                    else:
                        start = len(result) - 1 if start is None else start
                        stop = -1 if stop is None else stop
                    try:
                        index = range(start, stop, step if step is not None else 1)
                    except TypeError:
                        raise TypeError("Slice indices must be integers or None or have an __index__ method")
                return [get_item_recursive(result[i], indices[1:]) for i in index]
            else:
                raise TypeError("Valid indices are integers, slices (`:`), lists, tuples, ranges, and BoolArray")

        if isinstance(item, (int, slice)):
            return self.data[item]
        if isinstance(item, tuple):
            if len(item) > len(self.shape):
                raise IndexError("Too many indices for array: array is {}-dimensional, but {} were indexed".format(
                    len(self.shape), len(item)))
            else:
                return get_item_recursive(self.data, item)
        if isinstance(item, list):
            if all([isinstance(idx, int) for idx in item]):
                try:
                    return [self.data[idx] for idx in item]
                except IndexError:
                    raise IndexError("Valid indices are from -{} to {}".format(len(self.data), len(self.data) - 1))
        if isinstance(item, BoolArray):
            if self.shape != item.shape:
                raise MatchError
            return [value for value, flag in zip(self.flatten(), item.flatten()) if flag]
        raise TypeError("Valid indices are integers, slices (`:`), lists, tuples, ranges, and BoolArray")

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __round__(self, n=None):
        from .ufuncs import base_ufunc
        return base_ufunc(self, func=lambda number, digits: round(number.real, digits) + round(
            number.imag, digits) * 1j if isinstance(number, complex) else round(number, digits), args=[n])

    def __hash__(self):
        return hash(repr(self.data))

    def __add__(self, other):
        if isinstance(other, Array):
            if self.shape != other.shape:
                raise MatchError
            return fill(self.shape, [t1 + t2 for t1, t2 in zip(self.flatten(), other.flatten())], rtype=type(self))
        elif isinstance(other, (int, float, complex)):
            from .ufuncs import add
            return add(self, other)
        else:
            raise ValueError("Another must be an array or number")

    def __sub__(self, other):
        if isinstance(other, Array):
            if self.shape != other.shape:
                raise MatchError
            return fill(self.shape, [t1 - t2 for t1, t2 in zip(self.flatten(), other.flatten())], rtype=type(self))
        elif isinstance(other, (int, float, complex)):
            from .ufuncs import subtract
            return subtract(self, other)
        else:
            raise ValueError("Another must be an array or number")

    def __mul__(self, other):
        if isinstance(other, Array):
            if self.shape != other.shape:
                raise MatchError
            return fill(self.shape, [t1 * t2 for t1, t2 in zip(self.flatten(), other.flatten())], rtype=type(self))
        elif isinstance(other, (int, float, complex)):
            from .ufuncs import multiply
            return multiply(self, other)
        else:
            raise ValueError("Another must be an array or number")

    def __truediv__(self, other):
        if isinstance(other, Array):
            if self.shape != other.shape:
                raise MatchError
            return fill(self.shape, [t1 / t2 for t1, t2 in zip(self.flatten(), other.flatten())], rtype=type(self))
        elif isinstance(other, (int, float, complex)):
            from .ufuncs import divide
            return divide(self, other)
        else:
            raise ValueError("Another must be an array or number")

    def __floordiv__(self, other):
        if isinstance(other, Array):
            if self.shape != other.shape:
                raise MatchError
            return fill(self.shape, [t1 // t2 for t1, t2 in zip(self.flatten(), other.flatten())], rtype=type(self))
        elif isinstance(other, (int, float, complex)):
            from .ufuncs import floor_divide
            return floor_divide(self, other)
        else:
            raise ValueError("Another must be an array or number")

    def __mod__(self, other):
        if isinstance(other, Array):
            if self.shape != other.shape:
                raise MatchError
            return fill(self.shape, [t1 % t2 for t1, t2 in zip(self.flatten(), other.flatten())], rtype=type(self))
        elif isinstance(other, (int, float, complex)):
            from .ufuncs import modulo
            return modulo(self, other)
        else:
            raise ValueError("Another must be an array or number")

    def __pow__(self, _exp, _mod=None):
        if isinstance(_exp, (int, float, complex, Array)) and isinstance(
                _mod, (int, float, complex, Array, type(None))):
            from .ufuncs import power
            return power(self, _exp, _mod)
        else:
            raise ValueError("Exponential and modulus must both be arrays or numbers")

    __radd__ = __add__
    __rsub__ = __sub__
    __rmul__ = __mul__
    __rtruediv__ = __truediv__
    __rfloordiv__ = __floordiv__
    __rmod__ = __mod__
    __rpow__ = __pow__

    def __gt__(self, other):
        return self.comparison(other, lambda x, y: x > y)

    def __ge__(self, other):
        return self.comparison(other, lambda x, y: x >= y)

    def __eq__(self, other):
        return self.comparison(other, lambda x, y: x == y)

    def __le__(self, other):
        return self.comparison(other, lambda x, y: x <= y)

    def __lt__(self, other):
        return self.comparison(other, lambda x, y: x < y)

    def __ne__(self, other):
        return self.comparison(other, lambda x, y: x != y)

    def comparison(self, other, func):
        from .ufuncs import apply, base_ufunc
        if isinstance(other, (int, float, complex)):
            return apply(self, lambda x: func(x, other), BoolArray)
        elif isinstance(other, Array):
            if self.shape != other.shape:
                raise ValueError("Shapes must be equal for element-wise comparison")
            return base_ufunc(self, other, func=func, rtype=BoolArray)
        else:
            raise TypeError("Unsupported operand type(s) for comparison: '{}' and '{}'".format(type(self), type(other)))

    def apply(self, func):
        from .ufuncs import apply
        return apply(self, func)

    def flatten(self):
        data = self.data
        while isinstance(data[0], list):
            data = sum(data, [])
        return data

    def reshape(self, shape, repeat=True, pad=0):
        return fill(shape, self.flatten(), repeat, pad, type(self))

    def copy(self):
        from copy import deepcopy
        return deepcopy(self)

    def basic(self, func, axis=None):
        if axis is None:
            return func(self.flatten())

        def inner(a, flag):
            if flag:
                return [inner(_, flag - 1) for _ in a]
            elif all([isinstance(_, (int, float, complex)) for _ in a]):
                return func(a)
            else:
                return [inner(list(_), 0) for _ in zip(*a)]

        return type(self)(inner(self, axis))

    def sum(self, axis=None):
        return self.basic(sum, axis)

    def max(self, axis=None):
        return self.basic(max, axis)

    def min(self, axis=None):
        return self.basic(min, axis)

    def argmax(self, axis=None):
        return self.basic(lambda a: a.index(max(a)), axis)

    def argmin(self, axis=None):
        return self.basic(lambda a: a.index(min(a)), axis)

    def mean(self, axis=None):
        from .maths import mean
        return self.basic(mean, axis)

    def var(self, axis=None):
        from .maths import var
        return self.basic(var, axis)

    def std(self, axis=None):
        from .maths import std
        return self.basic(std, axis)

    def ptp(self, axis=None):
        from .maths import ptp
        return self.basic(ptp, axis)

    def median(self, axis=None):
        from .maths import median
        return self.basic(median, axis)

    def mode(self, axis=None):
        from .maths import mode
        return self.basic(mode, axis)

    def product(self, axis=None):
        from .maths import product
        return self.basic(product, axis)


def get_shape(data):
    _shape = []
    _sub = data
    while isinstance(_sub, list):
        _shape.append(len(_sub))
        _sub = _sub[0]
    return tuple(_shape)


def is_valid_array(_array, _shape):
    def inner(a, s, al, sl):
        if sl == 1:
            if not isinstance(a, list) or al != s[0]:
                raise ArrayError
        elif al == s[0]:
            for i in a:
                if not isinstance(i, list):
                    raise ArrayError
                inner(i, s[1:], len(i), sl - 1)
        else:
            raise ArrayError

    shape_len = len(_shape)
    if shape_len != 1 and not isinstance(_array, list):
        raise ArrayError
    inner(_array, _shape, len(_array), shape_len)


def array(data=None):
    return Array(data)


def aslist(data):
    if isinstance(data, list):
        return data
    if isinstance(data, Array):
        return data.data
    if isinstance(data, (tuple, str)):
        return list(data)
    if isinstance(data, set):
        return sorted(data)
    raise TypeError("Unable to convert to array type")


def asarray(data):
    return Array(aslist(data))


def full(shape: arr, fill_value: Any, rtype: Callable = Array) -> Any:
    def inner(data, length):
        return fill_value if length == 0 else [inner(data[1:], length - 1) for _ in range(data[0])]

    if isinstance(fill_value, list):
        raise TypeError("The filled value cannot be a list")
    result = inner(shape, len(shape))
    return result if rtype is list else rtype(result)


def full_like(a: Any, fill_value: Any, rtype: Callable = Array) -> Any:
    def inner(data):
        return [inner(item) for item in data] if isinstance(data, list) else fill_value

    if isinstance(fill_value, list):
        raise TypeError("The filled value cannot be a list")
    if isinstance(a, Array):
        a = a.data
    result = inner(a)
    return result if rtype is list else rtype(result)


def zeros(shape: arr, rtype: Callable = Array) -> Any:
    return full(shape, 0, rtype)


def zeros_like(a: Any, rtype: Callable = Array) -> Any:
    return full_like(a, 0, rtype)


def ones(shape: arr, rtype: Callable = Array) -> Any:
    return full(shape, 1, rtype)


def ones_like(a: Any, rtype: Callable = Array) -> Any:
    return full_like(a, 1, rtype)


def fill(shape: arr, sequence: arr = None, repeat: bool = True, pad: Any = 0, rtype: Callable = Array) -> Any:
    pointer = -1
    length = 1
    for item in shape:
        length *= item
    if sequence is None:
        sequence = list(range(length))
    total = len(sequence)
    last = total - 1

    def inner(_shape, depth):
        nonlocal pointer
        if depth == 0:
            if pointer == last and not repeat:
                return pad
            pointer += 1
            return sequence[pointer % total]
        else:
            return [inner(_shape[1:], depth - 1) for _ in range(_shape[0])]

    result = inner(shape, len(shape))
    return result if rtype is list else rtype(result)


def tensorproduct(*tensors: Array) -> Array:
    if not all([isinstance(tensor, Array) for tensor in tensors]):
        raise TypeError("All inputs must be tensors")
    if len(tensors) == 1:
        return tensors[0].copy()

    def mul(a, b):
        flattened_a = a.flatten()
        flattened_b = b.flatten()
        return Array(fill(a.shape + b.shape, [i * j for i in flattened_a for j in flattened_b], rtype=list), False)

    first = tensors[0]
    for second in tensors[1:]:
        first = mul(first, second)
    return first


class BoolArray(Array):
    def __init__(self, data=None, check=True):
        from .ufuncs import apply
        super().__init__(apply(Array(data, check), bool, list), False)

    def logic_op(self, other, func):
        from .ufuncs import ufunc_helper
        if isinstance(other, Array) and not isinstance(other, BoolArray) and not isinstance(other, bool):
            raise TypeError("Other must be a BoolArray or a scalar boolean value")
        return ufunc_helper(self, other, func)

    def __and__(self, other):
        return self.logic_op(other, lambda a, b: a and b)

    def __or__(self, other):
        return self.logic_op(other, lambda a, b: a or b)

    def __xor__(self, other):
        return self.logic_op(other, lambda a, b: a != b)

    def __invert__(self):
        return self.apply(lambda x: not x)


def boolarray(data=None):
    return BoolArray(data)



from .arrays import Array, fill
from .types import ShapeError

MatchError = ShapeError("The shapes of two tensors do not match")


class Tensor(Array):
    """
    It is a tensor class that supports basic operations.
    :param data: An array in the form of a list
    :param check: Check the rationality of the input array
    """

    def __init__(self, data, check=True):
        from warnings import warn
        warn("The 'Tensor' class is deprecated and will be removed in a future version. "
             "Please use the 'Array' class instead for similar functionality.", FutureWarning)
        super().__init__(data, check)

    def __add__(self, other):
        if isinstance(other, Tensor):
            if self.shape != other.shape:
                raise MatchError
            return Tensor(fill(self.shape, [t1 + t2 for t1, t2 in zip(self.flatten(), other.flatten())], rtype=list),
                          False)
        elif isinstance(other, (int, float, complex)):
            return Tensor(tensor_and_number(self, "+", other), False)
        else:
            raise ValueError("The other must be a tensor or a number")

    def __sub__(self, other):
        if isinstance(other, Tensor):
            if self.shape != other.shape:
                raise MatchError
            return Tensor(fill(self.shape, [t1 - t2 for t1, t2 in zip(self.flatten(), other.flatten())], rtype=list),
                          False)
        elif isinstance(other, (int, float, complex)):
            return Tensor(tensor_and_number(self, "-", other), False)
        else:
            raise ValueError("The other must be a tensor or a number")

    def __mul__(self, other):
        if isinstance(other, Tensor):
            if self.shape != other.shape:
                raise MatchError
            return Tensor(fill(self.shape, [t1 * t2 for t1, t2 in zip(self.flatten(), other.flatten())], rtype=list),
                          False)
        elif isinstance(other, (int, float, complex)):
            return Tensor(tensor_and_number(self, "*", other), False)
        else:
            raise ValueError("The other must be a tensor or a number")

    def __matmul__(self, other):
        if isinstance(other, Tensor):
            if self.shape[:-2] != other.shape[:-2]:
                raise MatchError

            def matmul(s, o):
                if not isinstance(s[0][0], list) or not isinstance(o[0][0], list):
                    return [[sum([s[i][k] * o[k][j] for k in range(len(s[0]))])
                             for j in range(len(o[0]))] for i in range(len(s))]
                if len(s) != len(o):
                    raise ValueError("Tensor dimensions do not match")
                return [matmul(t1, t2) for t1, t2 in zip(s, o)]

            return Tensor(matmul(self.data, other.data), False)
        else:
            raise ValueError("The other must be a tensor")


def tensor_and_number(tensor, operator, number):
    if isinstance(tensor, Tensor) or isinstance(tensor, list):
        _result = []
        for item in tensor:
            _result.append(tensor_and_number(item, operator, number))
        return _result
    else:
        if operator in ["+", "-", "*"]:
            return eval("{} {} {}".format(tensor, operator, number))


def ten(data: list) -> Tensor:
    return Tensor(data)


del Array



from .types import Any, Callable, Union, arr, ite, real


def frange(start: real, stop: real, step: float = 1.0) -> list:
    """
    Introduction
    ==========
    Float range

    Example
    ==========
    >>> frange(0, 10, 1.5)
    [0.0, 1.5, 3.0, 4.5, 6.0, 7.5, 9.0]
    >>>
    :param start: integer | float
    :param stop: integer | float
    :param step: float
    :return:
    """
    if isinstance(step, int):
        step = float(step)
    result = [_ * step for _ in range(int(start / step), int(stop / step + 1))]
    return result


def linspace(start: real, stop: real, number: int) -> list:
    """
    Introduction
    ==========
    Linear space

    Example
    ==========
    >>> linspace(2, 3, 4)
    [2.0, 2.3333333333333335, 2.6666666666666665, 3.0]
    >>>
    :param start: integer | float
    :param stop: integer | float
    :param number: integer
    :return:
    """
    if number == 1:
        return [start]
    elif number == 2:
        return [start, stop]
    else:
        step = (stop - start) / (number - 1)
        return [start + i * step for i in range(number)]


def geomspace(start: real, stop: real, number: int) -> list:
    """
    Introduction
    ==========
    Geometric space

    Example
    ==========
    >>> geomspace(2, 3, 4)
    [2.0, 2.2894284851066637, 2.620741394208897, 3.0]
    >>>
    :param start: integer | float
    :param stop: integer | float
    :param number: integer
    :return:
    """
    if number == 1:
        return [start]
    elif number == 2:
        return [start, stop]
    else:
        step = (stop / start) ** (1 / (number - 1))
        return [start * step ** i for i in range(number)]


def dedup(iterable: ite) -> ite:
    """
    Introduction
    ==========
    Data deduplication

    Example
    ==========
    >>> dedup(["T", "h", "i", "s", " ", "i", "s", " ", "a", " ", "l", "i", "s", "t"])
    ['T', 'h', 'i', 's', ' ', 'a', 'l', 't']
    >>>
    :param iterable: list | tuple | string
    :return:
    """
    result = []
    for item in iterable:
        if item not in result:
            result.append(item)
    if isinstance(iterable, list):
        return result
    elif isinstance(iterable, tuple):
        return tuple(result)
    elif isinstance(iterable, str):
        return "".join(result)
    else:
        raise TypeError("Iterable can only be a list, tuple or str")


def classify(array: arr) -> dict:
    """
    Introduction
    ==========
    Data classification

    Example
    ==========
    >>> classify((1, 2.3, 4 + 5j, 6 - 7j, 8.9, 0))
    {<class 'int'>: [1, 0], <class 'float'>: [2.3, 8.9], <class 'complex'>: [(4+5j), (6-7j)]}
    >>>
    :param array: list | tuple
    :return:
    """
    result = {}
    for item in array:
        _type = type(item)
        if _type in result:
            result[_type].append(item)
        else:
            result[_type] = [item]
    return result


def split(iterable: ite, key: arr, retain: bool = False) -> list:
    """
    Introduction
    ==========
    Data splitting

    Example
    ==========
    >>> split((1, 2, 2, 3, 3, 2, 2, 1, 2, 3, 4, 5, 6, 7, 8), [1, 3, 5], retain=True)
    [(), (1,), (2, 2), (3,), (), (3,), (2, 2), (1,), (2,), (3,), (4,), (5,), (6, 7, 8)]
    >>>
    :param iterable: list | tuple
    :param key: list | tuple
    :param retain: bool
    :return:
    """
    key = dedup(key)
    if not isinstance(retain, bool):
        raise TypeError("Only Boolean values can be used to determine whether to retain the key")
    if not isinstance(key, (list, tuple)):
        raise TypeError("The parameter 'key' can only be a list or tuple")
    result = []
    pointer = 0
    if isinstance(iterable, str):
        if key == [""]:
            raise ValueError("When iterable is a string, the key cannot be a list of empty strings")
        while True:
            indexes = {}
            for k in key:
                index = iterable.find(k, pointer)
                if index != -1 and index not in indexes:
                    indexes[index] = k
            if indexes:
                index = min(indexes)
                result.append(iterable[pointer:index])
                pointer = index
                if retain is True:
                    result.append(indexes[index])
                pointer += len(indexes[index])
            else:
                break
        result.append(iterable[pointer:])
    elif isinstance(iterable, (list, tuple)):
        for item in range(len(iterable)):
            if iterable[item] in key:
                result.append(iterable[pointer:item])
                if isinstance(iterable, (list, tuple)):
                    pointer = item + 1
                if retain is True:
                    result.append(iterable[item:item + 1])
        result.append(iterable[pointer:])
    else:
        raise TypeError("Iterable can only be a list, tuple or str")
    return result


def primality(n: int, iter_num: int = 10) -> bool:
    """
    Introduction
    ==========
    This function checks if a number is a prime using the Miller-Rabin primality test,
    which is a probabilistic test to determine if a number is a probable prime.

    Example
    ==========
    >>> primality(2 ** 4423 - 1)
    True
    >>>
    :param n: The number to be tested for primality.
    :param iter_num: The number of iterations for the test, which affects the accuracy.
        More iterations reduce the probability of falsely identifying a composite number as prime.
    :return: True if the number is a probable prime, False otherwise.
    """
    from random import randint
    if n == 2:
        return True
    elif n & 1 == 0 or n < 2:
        return False
    m, s = n - 1, 0
    while m & 1 == 0:
        m >>= 1
        s += 1
    for _ in range(iter_num):
        b = pow(randint(2, n - 1), m, n)
        if b == 1 or b == n - 1:
            continue
        for __ in range(s - 1):
            b = b * b % n
            if b == n - 1:
                break
        else:
            return False
    return True


def primes(limit: int) -> list:
    """
    Introduction
    ==========
    This function generates a list of prime numbers up to a specified limit.

    Example
    ==========
    >>> primes(100)
    [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]
    >>>
    :param limit: The upper limit to which prime numbers are to be found (inclusive).
    :return: A list of prime numbers up to the specified limit.
    """
    result = list(range(limit + 1))
    result[1] = 0
    p = 2
    s = 4
    while s <= limit:
        if result[p]:
            for i in range(s, limit + 1, p):
                result[i] = 0
        p += 1
        s = p * p
    return [x for x in result if x != 0]


def semiprimes(limit: int) -> list:
    """
    Introduction
    ==========
    This function generates a list of semiprime numbers up to a specified limit.
    A semiprime is a natural number that is the product of two prime numbers.

    Example
    ==========
    >>> semiprimes(64)
    [4, 6, 9, 10, 14, 15, 21, 22, 25, 26, 33, 34, 35, 38, 39, 46, 49, 51, 55, 57, 58, 62]
    >>>
    :param limit: The upper limit to which semiprime numbers are to be found (inclusive).
    :return: A list of semiprime numbers up to the specified limit.
    """
    ps = primes(limit // 2)
    length = len(ps)
    result = []
    for i in range(length):
        first = ps[i]
        maximum = limit // first
        for j in range(i, length):
            second = ps[j]
            if second > maximum:
                break
            result.append(first * second)
    return sorted(result)


def twinprimes(limit: int) -> list:
    """
    Introduction
    ==========
    This function generates a list of twin prime pairs up to a specified limit.
    Twin primes are a pair of prime numbers that have a difference of two.

    Example
    ==========
    >>> twinprimes(100)
    [(3, 5), (5, 7), (11, 13), (17, 19), (29, 31), (41, 43), (59, 61), (71, 73)]
    >>>
    :param limit: The upper limit to which twin prime pairs are to be found (inclusive).
    :return: A list of twin prime pairs up to the specified limit.
    """
    ps = primes(limit)
    return [(p, q) for p, q in zip(ps, ps[1:]) if q - p == 2]


def prime_factors(integer: int, dictionary: bool = False, pollard_rho: bool = True) -> Union[list, dict]:
    """
    Introduction
    ==========
    This function computes the prime factors of a given integer. It can return the factors
    as a list or a dictionary, and it can use the Pollard's rho algorithm for factorization.

    Example
    ==========
    >>> prime_factors(2305567963945518424753102147331756070)
    [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]
    >>>
    :param integer: The integer to be factorized.
    :param dictionary: A boolean indicating whether to return the factors as a dictionary with the prime factors as
        keys and their counts as values.
    :param pollard_rho: A boolean indicating whether to use Pollard's rho algorithm for factorization. If False,
        a simple trial division method is used.
    :return: A list of prime factors or a dictionary with prime factors and their counts, depending on the
        'dictionary' parameter.
    """
    from math import gcd
    from random import randint

    def normal(n):
        factors = []
        divisor = 2
        while n > 1:
            tmp = divmod(n, divisor)
            if tmp[1]:
                divisor += 1
            else:
                factors.append(divisor)
                n = tmp[0]
        return factors

    def pollard(x, c):
        i, k = 1, 2
        x0 = randint(0, x)
        y = x0
        while 1:
            i += 1
            x0 = (x0 * x0 % x + c) % x
            d = gcd(y - x0, x)
            if d != 1 and d != x:
                return d
            if y == x0:
                return x
            if i == k:
                y = x0
                k += k

    def recursion(n):
        result = []
        if primality(n):
            return [n]
        p = n
        while p >= n:
            p = pollard(p, randint(1, n - 1))
        result.extend(recursion(p))
        result.extend(recursion(n // p))
        return result

    if integer > 0:
        output = recursion(integer) if pollard_rho else normal(integer)
        if dictionary:
            return {number: output.count(number) for number in set(output)}
        else:
            return sorted(output)
    else:
        return {} if dictionary else []


def magic_square(n: int) -> list:
    """
    Introduction
    ==========
    Generate a magic square of a specified order

    Example
    ==========
    >>> magic_square(4)
    [[16, 2, 3, 13], [5, 11, 10, 8], [9, 7, 6, 12], [4, 14, 15, 1]]
    >>>
    :param n: integer
    :return:
    """
    if not isinstance(n, int) or n < 3:
        raise ValueError("The order of the magic square must be an integer greater than or equal to three")
    result = [[0] * n for _ in range(n)]

    def odd():
        i, j = 0, n // 2
        for m in range(1, n * n + 1):
            result[i][j] = m
            new_i, new_j = (i - 1) % n, (j + 1) % n
            if result[new_i][new_j] != 0:
                new_i, new_j = (i + 1) % n, j
            i, j = new_i, new_j

    def single_even():
        half = n // 2
        left = magic_square(half)
        x = n ** 2 // 4
        k = (n - 2) // 4
        for i in range(half):
            for j in range(half):
                result[i][j] = left[i][j]
                result[i][j + half] = left[i][j] + 2 * x
                result[i + half][j] = left[i][j] + 3 * x
                result[i + half][j + half] = left[i][j] + x
        for i in range(half):
            for j in range(n):
                if j > (half + k + 1):
                    result[i][j], result[i + half][j] = result[i + half][j], result[i][j]
                if j < k:
                    result[i][j], result[i + half][j] = result[i + half][j], result[i][j]
        result[k][0], result[k + half][0] = result[k + half][0], result[k][0]
        result[k][k], result[k + half][k] = result[k + half][k], result[k][k]

    def double_even():
        c1 = 1
        c2 = n * n
        for i in range(n):
            for j in range(n):
                if i % 4 == j % 4 or (i + j) % 4 == 3:
                    result[i][j] = c2
                else:
                    result[i][j] = c1
                c2 -= 1
                c1 += 1

    if n & 1:
        odd()
    elif n & 3:
        single_even()
    else:
        double_even()
    return result


def levenshtein(x: ite, y: ite) -> int:
    """
    Introduction
    ==========
    Calculate the Levenshtein distance between two sequences.

    The Levenshtein distance is a measure of the difference between two sequences. It is defined as the minimum number
    of single-character edits (i.e., insertions, deletions, or substitutions) required to change one sequence into the
    other. This function supports any type of iterable sequences, such as strings, lists, or tuples.

    Example
    ==========
    >>> levenshtein("kitten", "sitting")
    3
    >>>
    :param x: First sequence to compare.
    :param y: Second sequence to compare.
    :return: The Levenshtein distance between the two sequences.
    """
    if len(x) > len(y):
        x, y = y, x
    previous_row = list(range(len(x) + 1))
    for i2, c2 in enumerate(y):
        current_row = [i2 + 1]
        for i1, c1 in enumerate(x):
            insertions = previous_row[i1 + 1] + 1
            deletions = current_row[i1] + 1
            substitutions = previous_row[i1] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row
    return previous_row[-1]


def lcsubseq(x: ite, y: ite) -> list:
    """
    Introduction
    ==========
    Calculate the Longest Common Subsequence (LCS) between two sequences.

    The Longest Common Subsequence is a sequence that appears in both given sequences in the same order, but not
    necessarily consecutively. This function uses dynamic programming to build a table that stores the lengths of the
    longest common subsequences of substring pairs, and then backtracks to construct the LCS.

    Example
    ==========
    >>> lcsubseq("ABCBDAB", "BDCAB")
    ['B', 'D', 'A', 'B']
    >>>
    :param x: First sequence to compare.
    :param y: Second sequence to compare.
    :return: The Longest Common Subsequence between the two sequences.
    """
    m, n = len(x), len(y)
    dp = [[0] * (n + 1) for _ in range(m + 1)]
    for i in range(1, m + 1):
        for j in range(1, n + 1):
            if x[i - 1] == y[j - 1]:
                dp[i][j] = dp[i - 1][j - 1] + 1
            else:
                dp[i][j] = max(dp[i - 1][j], dp[i][j - 1])
    i, j = m, n
    seq = []
    while i > 0 and j > 0:
        if x[i - 1] == y[j - 1]:
            seq.append(x[i - 1])
            i -= 1
            j -= 1
        elif dp[i - 1][j] > dp[i][j - 1]:
            i -= 1
        else:
            j -= 1
    seq.reverse()
    return seq


def lcsubstr(x: ite, y: ite) -> list:
    """
    Introduction
    ==========
    Calculate the Longest Common Substring between two sequences.

    The Longest Common Substring is a substring that appears in both given sequences in the same order and is
    consecutive. This function uses dynamic programming to build a table that stores the lengths of the longest common
    substrings of substring pairs, and keeps track of the maximum length and its ending index to extract the substring.

    Example
    ==========
    >>> lcsubstr("ABCBDAB", "BDCAB")
    ['A', 'B']
    >>>
    :param x: First sequence to compare.
    :param y: Second sequence to compare.
    :return: The Longest Common Substring between the two sequences.
    """
    m, n = len(x), len(y)
    dp = [[0] * (n + 1) for _ in range(m + 1)]
    max_len = 0
    end_idx = 0
    for i in range(1, m + 1):
        for j in range(1, n + 1):
            if x[i - 1] == y[j - 1]:
                dp[i][j] = dp[i - 1][j - 1] + 1
                if dp[i][j] > max_len:
                    max_len = dp[i][j]
                    end_idx = i
            else:
                dp[i][j] = 0
    return list(x[end_idx - max_len:end_idx])


def cos_sim(seq1: ite, seq2: ite, is_vector: bool = False) -> float:
    """
    Introduction
    ==========
    Calculate the cosine similarity between two sequences.

    The cosine similarity is a measure of similarity between two non-zero vectors. It is defined as the cosine of the
    angle between them, which is computed as the dot product of the vectors divided by the product of their magnitudes.
    This function supports both numerical vectors and frequency distributions of sequences.

    Example
    ==========
    >>> cos_sim("hello world", "world hello")
    0.9999999999999998
    >>>
    :param seq1: First sequence to compare.
    :param seq2: Second sequence to compare.
    :param is_vector: A boolean indicating whether the input sequences are numerical vectors. Default is False.
    :return: The cosine similarity between the two sequences, ranging from -1 to 1.
    """
    from .maths import freq
    if is_vector:
        dot_product = sum([a * b for a, b in zip(seq1, seq2)])
        magnitude_seq1 = sum([a ** 2 for a in seq1]) ** 0.5
        magnitude_seq2 = sum([b ** 2 for b in seq2]) ** 0.5
    else:
        freq1 = freq(seq1)
        freq2 = freq(seq2)
        dot_product = sum([freq1.get(k, 0) * freq2.get(k, 0) for k in set(freq1) | set(freq2)])
        magnitude_seq1 = sum([v ** 2 for v in freq1.values()]) ** 0.5
        magnitude_seq2 = sum([v ** 2 for v in freq2.values()]) ** 0.5
    if magnitude_seq1 * magnitude_seq2 == 0:
        return 0.0
    return dot_product / (magnitude_seq1 * magnitude_seq2)


def kmp_table(pattern: ite) -> list:
    """
    Introduction
    ==========
    Generate the KMP (Knuth-Morris-Pratt) table for a given pattern.

    The KMP table is used to efficiently find occurrences of a pattern within a sequence by avoiding unnecessary
    comparisons after a mismatch. This table determines how many characters can be skipped after a mismatch.

    Example
    ========
    >>> kmp_table("AGCTGATCGTACGTAAGCTAGCTA")
    [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 2, 3, 4, 1, 2, 3, 4, 1]
    >>>
    :param pattern: The pattern for which to generate the KMP table.
    :return: A list representing the KMP table for the given pattern.
    """
    pattern_len = len(pattern)
    table = [0] * pattern_len
    pos, cnd = 1, 0
    while pos < pattern_len:
        if pattern[pos] == pattern[cnd]:
            cnd += 1
            table[pos] = cnd
            pos += 1
        elif cnd > 0:
            cnd = table[cnd - 1]
        else:
            table[pos] = 0
            pos += 1
    return table


def findall(seq: ite, pat: ite) -> list:
    """
    Introduction
    ==========
    Find all indices of the subsequence 'pat' in 'seq'.

    This function is designed to handle sequences such as lists, tuples, or strings and find all indices
    of specified subsequences. It allows overlapping matches.

    Example
    ========
    >>> findall([2, 1, 2, 1, 2, 1, 2, 1], [1, 2, 1, 2])
    [1, 3]
    >>>
    :param seq: The sequence in which to find the subsequence.
    :param pat: The subsequence to be found.
    :return: A list of starting indices where the subsequence is found.
    """
    if not isinstance(seq, (list, tuple, str)):
        raise TypeError("Parameter 'seq' must be a list, tuple, or string, got {}".format(type(seq)))
    if not isinstance(pat, (list, tuple, str)):
        raise TypeError("Parameter 'pat' must be a list, tuple, or string, got {}".format(type(pat)))
    if not pat:
        raise ValueError("Parameter 'pat' must not be empty")
    if isinstance(seq, str):
        seq = tuple(seq)
    if isinstance(pat, str):
        pat = tuple(pat)
    pat_len = len(pat)
    table = kmp_table(pat)
    indices = []
    i, j = 0, 0
    seq_len = len(seq)
    while i < seq_len:
        if j == pat_len:
            indices.append(i - j)
            j = table[j - 1]
        elif seq[i] == pat[j]:
            i += 1
            j += 1
        else:
            if j != 0:
                j = table[j - 1]
            else:
                i += 1
    if j == pat_len:
        indices.append(i - j)
    return indices


def replace(seq: arr, old: arr, new: arr, count: int = -1) -> arr:
    """
    Introduction
    ==========
    Replace occurrences of the subsequence 'old' in 'seq' with 'new'.

    This function is designed to handle sequences such as lists or tuples and replace specified subsequences
    with new ones. It also allows limiting the number of replacements.

    Example
    ==========
    >>> replace([1, 2, 3, 4, 2, 3], [2, 3], [5, 6])
    [1, 5, 6, 4, 5, 6]
    >>>
    :param seq: The sequence in which to replace the subsequence.
    :param old: The subsequence to be replaced.
    :param new: The subsequence to replace with.
    :param count: The maximum number of replacements to perform. Default is -1 (unlimited).
    :return: The modified sequence with replacements.
    """
    if not isinstance(seq, (list, tuple)):
        raise TypeError("Parameter 'seq' must be a list or a tuple, got {}".format(type(seq)))
    if not isinstance(old, (list, tuple)):
        raise TypeError("Parameter 'old' must be a list or a tuple, got {}".format(type(old)))
    if not isinstance(new, (list, tuple)):
        raise TypeError("Parameter 'new' must be a list or a tuple, got {}".format(type(new)))
    if not old:
        raise ValueError("Parameter 'old' must not be empty")
    if not isinstance(count, int) or count < 0 and count != -1:
        raise ValueError("Parameter 'count' must be a non-negative integer or -1, got {}".format(count))
    if count == 0:
        return seq[:]
    old_len = len(old)
    table = kmp_table(old)
    result = []
    i, j = 0, 0
    replaced = 0
    seq_len = len(seq)
    while i < seq_len:
        if j == old_len:
            result.extend(new)
            replaced += 1
            j = 0
            if count != -1 and replaced == count:
                result.extend(seq[i:])
                break
        elif seq[i] == old[j]:
            i += 1
            j += 1
        else:
            if j != 0:
                result.extend(old[:j])
                j = table[j - 1]
            else:
                result.append(seq[i])
                i += 1
    if j == old_len:
        result.extend(new)
    elif j > 0:
        result.extend(old[:j])
    return tuple(result) if isinstance(seq, tuple) else result


def fast_pow(a: Any, n: int, init: Any = 1, mul: Callable = None) -> Any:
    """
    Introduction
    ==========
    This function computes the power of a given base using the fast power algorithm.

    It is a versatile tool that can be used with any data type supporting multiplication, including the ability to
    perform matrix exponentiation by setting an initial matrix and a matrix multiplication function.
    For instance, to compute the matrix power of a 2x2 matrix, you can define the matrix as the base 'a',
    set an identity matrix as the 'init', and provide a function for matrix multiplication as 'mul'.

    Example
    ==========
    >>> fast_pow([[1, 2], [3, 4]], 31, init=[[1, 0], [0, 1]], mul=lambda _a, _b:
    ... [[_a[0][0] * _b[0][0] + _a[0][1] * _b[1][0], _a[0][0] * _b[0][1] + _a[0][1] * _b[1][1]],
    ... [_a[1][0] * _b[0][0] + _a[1][1] * _b[1][0], _a[1][0] * _b[0][1] + _a[1][1] * _b[1][1]]])
    [[10306408007805049875493, 15020838414172076413286], [22531257621258114619929, 32837665629063164495422]]
    >>>
    :param a: The base of the exponentiation, which can be any type of data.
    :param n: The exponent, which must be a non-negative integer.
    :param init: The initial value, which defaults to 1. It is used as the starting point for the computation.
    :param mul: A custom multiplication operation, which defaults to None. If provided, it should be a callable
                that takes two arguments and returns their product.
    :return: The result of the exponentiation, which will be of the same type as the 'init' parameter.
    """
    if mul is None:
        def mul(_a, _b):
            return _a * _b
    res = init
    while n:
        if n & 1:
            res = mul(res, a)
        a = mul(a, a)
        n >>= 1
    return res


if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=True)



def mediantest(*samples, ties="below", lambda_=1, corr=True):
    from math import isnan
    from .matrices import mat
    from .maths import median
    if len(samples) < 2:
        raise ValueError("median_test requires two or more samples.")
    data = [list(sample) for sample in samples]
    cdata = [item for sublist in data for item in sublist]
    grand_median = median([x for x in cdata if x is not None and not (isinstance(x, float) and isnan(x))])
    table = [[0] * len(data) for _ in range(2)]
    for k, sample in enumerate(data):
        above = sum([x > grand_median for x in sample if x is not None
                     and not (isinstance(x, float) and isnan(x))])
        below = sum([x < grand_median for x in sample if x is not None
                     and not (isinstance(x, float) and isnan(x))])
        equal = len(sample) - (above + below)
        table[0][k] += above
        table[1][k] += below
        if ties == "below":
            table[1][k] += equal
        elif ties == "above":
            table[0][k] += equal
        elif ties != "ignore":
            raise ValueError("Invalid ties option: {}. Valid options are: ['below', 'above', 'ignore']".format(ties))
    stat, p, _, _ = chi2_cont(table, lambda_=lambda_, corr=corr)
    return stat, p, grand_median, mat(table)


def skewtest(data: list, two_tailed: bool = True) -> tuple:
    from math import log, sqrt
    from .dists import normal_cdf
    from .maths import skew
    n = len(data)
    y = skew(data) * sqrt(((n + 1) * (n + 3)) / (6 * (n - 2)))
    beta = 3 * (n ** 2 + 27 * n - 70) * (n + 1) * (n + 3) / ((n - 2) * (n + 5) * (n + 7) * (n + 9))
    w = sqrt(2 * (beta - 1)) - 1
    delta = 1 / sqrt(0.5 * log(w))
    alpha = sqrt(2 / (w - 1))
    z = delta * log(y / alpha + sqrt((y / alpha) ** 2 + 1))
    p_value = normal_cdf(z) if z > 0 else 1 - normal_cdf(z)
    if two_tailed:
        p_value = 2 * min(p_value, 1 - p_value)
    return z, p_value


def kurttest(data: list, two_tailed: bool = True) -> tuple:
    from math import sqrt
    from .dists import normal_cdf
    from .maths import kurt, sign
    n = len(data)
    b = kurt(data, fisher=False)
    e = 3 * (n - 1) / (n + 1)
    vb = 24 * n * (n - 2) * (n - 3) / ((n + 1) * (n + 1) * (n + 3) * (n + 5))
    x = (b - e) / sqrt(vb)
    sb = 6 * (n * n - 5 * n + 2) / ((n + 7) * (n + 9)) * sqrt((6 * (n + 3) * (n + 5)) / (n * (n - 2) * (n - 3)))
    a = 6 + 8 / sb * (2 / sb + sqrt(1 + 4 / (sb ** 2)))
    d = 1 + x * sqrt(2 / (a - 4))
    t = sign(d) * (d == 0 and float("nan") or ((1 - 2 / a) / abs(d)) ** (1 / 3))
    z = (1 - 2 / (9 * a) - t) / sqrt(2 / (9 * a))
    p_value = normal_cdf(z) if z > 0 else 1 - normal_cdf(z)
    if two_tailed:
        p_value = 2 * min(p_value, 1 - p_value)
    return z, p_value


def normaltest(data: list) -> tuple:
    from .dists import chi2_cdf
    s = skewtest(data)[0]
    k = kurttest(data)[0]
    n = s * s + k * k
    return n, 1 - chi2_cdf(2, n)


def chisquare(observed: list, expected: list = None) -> tuple:
    from .dists import chi2_pdf
    from .maths import integ
    if expected is None:
        expected = [1 / len(observed)] * len(observed)
    chi2_stat = sum([(o - e) ** 2 / e for o, e in zip(observed, expected)])
    dof = len(observed) - 1
    try:
        p_value = integ(chi2_pdf, chi2_stat, 1000, df=dof)
    except OverflowError:
        p_value = float("nan")
    return chi2_stat, p_value


def chi2_cont(contingency: list, lambda_: float = 1.0, calc_p: bool = True, corr: bool = True) -> tuple:
    from .matrices import mat
    from .dists import chi2_cdf
    from .maths import xlogy
    contingency = mat(contingency)
    dof = (contingency.rows - 1) * (contingency.cols - 1)
    row_sums = [sum(row) for row in contingency]
    col_sums = [sum([contingency[i][j] for i in range(contingency.rows)]) for j in range(contingency.cols)]
    grand_total = sum(row_sums)
    expected = [[row_sums[i] * col_sums[j] / grand_total for j in range(contingency.cols)]
                for i in range(contingency.rows)]
    if dof == 1 and corr:
        corrected = [[0, 0], [0, 0]]
        for i in range(2):
            for j in range(2):
                diff = expected[i][j] - contingency[i][j]
                direction = 1 if diff > 0 else -1 if diff < 0 else 0
                magnitude = min(0.5, abs(diff))
                corrected[i][j] = contingency[i][j] + direction * magnitude
        contingency = mat(corrected)
    chi2 = 0
    for i in range(contingency.rows):
        for j in range(contingency.cols):
            observed_freq = contingency[i][j]
            expected_freq = expected[i][j]
            if expected_freq > 0:
                if lambda_ == 1:
                    terms = (observed_freq - expected_freq) ** 2 / expected_freq
                elif lambda_ == 0:
                    terms = 2.0 * xlogy(observed_freq, observed_freq / expected_freq)
                elif lambda_ == -1:
                    terms = 2.0 * xlogy(expected_freq, expected_freq / observed_freq)
                else:
                    terms = observed_freq * ((observed_freq / expected_freq) ** lambda_ - 1)
                    terms /= 0.5 * lambda_ * (lambda_ + 1)
                chi2 += terms
    try:
        p = 1 - chi2_cdf(chi2, df=dof) if calc_p else None
    except OverflowError:
        p = float("nan")
    return chi2, p, dof, mat(expected)



from math import cos as __cos, sin as __sin
from random import randint as __randint, random as __random, uniform as __uniform
from .types import real as __real


def randint_polar(left: int, right: int, mod: __real = None, angle: __real = None) -> complex:
    """
    Generate a complex number in polar form with integer modulus and angle
    randomly selected from a specified range.
    :param left: The lower bound (inclusive) of the random number range.
    :param right: The upper bound (inclusive) of the random number range.
    :param mod: The modulus of the complex number. If None, it is randomly
                selected from the range [left, right].
    :param angle: The angle of the complex number. If None, it is randomly
                  selected from the range [0, 2π].
    :return: A complex number in polar form.
    """
    mod = __randint(left, right) if mod is None else mod
    angle = __uniform(0, 6.283185307179586) if angle is None else angle
    return mod * (__cos(angle) + __sin(angle) * 1j)


def randint_rect(left: int, right: int, real: __real = None, imag: __real = None) -> complex:
    """
    Generate a complex number in rectangular form with integer real and
    imaginary parts randomly selected from a specified range.
    :param left: The lower bound (inclusive) of the random number range.
    :param right: The upper bound (inclusive) of the random number range.
    :param real: The real part of the complex number. If None, it is randomly
                 selected from the range [left, right].
    :param imag: The imaginary part of the complex number. If None, it is
                 randomly selected from the range [left, right].
    :return: A complex number in rectangular form.
    """
    real = __randint(left, right) if real is None else real
    imag = __randint(left, right) if imag is None else imag
    return complex(real, imag)


def random_polar(mod: __real = None, angle: __real = None) -> complex:
    """
    Generate a complex number in polar form with floating-point modulus and
    angle randomly selected from the range [0, 2π].
    :param mod: The modulus of the complex number. If None, it is randomly
                selected from the range [0, 1].
    :param angle: The angle of the complex number. If None, it is uniformly
                  selected from the range [0, 2π].
    :return: A complex number in polar form.
    """
    mod = __random() if mod is None else mod
    angle = __uniform(0, 6.283185307179586) if angle is None else angle
    return mod * (__cos(angle) + __sin(angle) * 1j)


def random_rect(real: __real = None, imag: __real = None) -> complex:
    """
    Generate a complex number in rectangular form with floating-point real and
    imaginary parts randomly selected from the range [0, 1].
    :param real: The real part of the complex number. If None, it is randomly
                 selected from the range [0, 1].
    :param imag: The imaginary part of the complex number. If None, it is
                 randomly selected from the range [0, 1].
    :return: A complex number in rectangular form.
    """
    real = __random() if real is None else real
    imag = __random() if imag is None else imag
    return complex(real, imag)


def uniform_polar(left: __real, right: __real, mod: __real = None, angle: __real = None) -> complex:
    """
    Generate a complex number in polar form with uniformly distributed
    floating-point modulus and angle from a specified range.
    :param left: The lower bound (inclusive) of the random number range.
    :param right: The upper bound (inclusive) of the random number range.
    :param mod: The modulus of the complex number. If None, it is uniformly
                selected from the range [left, right].
    :param angle: The angle of the complex number. If None, it is uniformly
                  selected from the range [0, 2π].
    :return: A complex number in polar form.
    """
    mod = __uniform(left, right) if mod is None else mod
    angle = __uniform(0, 6.283185307179586) if angle is None else angle
    return mod * (__cos(angle) + __sin(angle) * 1j)


def uniform_rect(left: __real, right: __real, real: __real = None, imag: __real = None) -> complex:
    """
    Generate a complex number in rectangular form with uniformly distributed
    floating-point real and imaginary parts from a specified range.
    :param left: The lower bound (inclusive) of the random number range.
    :param right: The upper bound (inclusive) of the random number range.
    :param real: The real part of the complex number. If None, it is uniformly
                 selected from the range [left, right].
    :param imag: The imaginary part of the complex number. If None, it is
                 uniformly selected from the range [left, right].
    :return: A complex number in rectangular form.
    """
    real = __uniform(left, right) if real is None else real
    imag = __uniform(left, right) if imag is None else imag
    return complex(real, imag)



import math
from .types import Union, arr, num, real, ite

__and = {0, 1, 4, 9, 16, 17, 25, 33, 36, 41, 49, 57, 64, 65, 68, 73, 81, 89, 97, 100, 105, 113,
         121, 129, 132, 137, 144, 145, 153, 161, 164, 169, 177, 185, 193, 196, 201, 209, 217, 225, 228, 233, 241, 249}
__mod = {0, 1, 4, 9, 16, 25, 36, 49, 64, 81, 100, 121, 144, 145, 160, 169,
         180, 196, 225, 241, 244, 256, 265, 289, 304, 324, 340, 361, 369, 385, 400, 409,
         436, 441, 481, 484, 496, 505, 529, 544, 576, 580, 585, 601, 625, 640, 649, 676}


def lower_gamma(s: num, x: num) -> num:
    """
    Introduction
    ==========
    Calculate the lower incomplete gamma function, denoted as γ(s, x) or P(s, x).

    Example
    ==========
    >>> lower_gamma(1, 2)
    0.8646647167633872
    >>>
    :param s: The shape parameter of the gamma function.
    :param x: The upper limit of the integration.
    :return: The lower incomplete gamma function of parameter s at x.
    """
    if x == 0:
        return x
    result = 0
    i = 0
    term = 1
    while term:
        try:
            term *= x / (s + i)
            temp = result
            result += term
            if result == temp:
                break
            i += 1
        except OverflowError:
            break
    return result * x ** (s - 1) * 2.718281828459045 ** (-x)


def upper_gamma(s: num, x: num) -> num:
    """
    Introduction
    ==========
    Calculate the upper incomplete gamma function, denoted as Q(s, x).

    Example
    ==========
    >>> upper_gamma(1, 2)
    0.1353352832366128
    >>>
    :param s: The shape parameter of the gamma function.
    :param x: The lower limit of the integration.
    :return: The upper incomplete gamma function of parameter s at x.
    """
    return math.gamma(s) - lower_gamma(s, x)


def sumprod(*arrays: arr) -> num:
    """
    Introduction
    ==========
    The sum of the products of the corresponding elements in multiple arrays

    Example
    ==========
    >>> sumprod([1, 2, 3], [4, 5, 6], [7, 8, 9])
    270.0
    >>>
    :param arrays: list | tuple
    :return:
    """
    return math.fsum(map(product, zip(*arrays)))


def xlogy(x: num, y: num) -> num:
    """
    Introduction
    ==========
    Compute the product of x and the logarithm of y for real and complex numbers.

    Example
    ==========
    >>> xlogy(2, 3)
    2.1972245773362196
    >>>
    :param x: Any real or complex number.
    :param y: A positive real number or a complex number with a positive real part,
        as logarithm is only defined for positive real values or complex numbers with a positive real part.
    :return: The result of x * log(y).
    """
    if isinstance(y, (int, float)) and y <= 0:
        raise ValueError("xlogy is not defined for y <= 0 when y is a real number")
    elif isinstance(y, complex) and y.real <= 0:
        raise ValueError("xlogy is not defined for y with a non-positive real part when y is a complex number")
    elif x == 0:
        return 0
    elif isinstance(x, (int, float)) and isinstance(y, (int, float)):
        return x * math.log(y)
    else:
        from cmath import log
        return x * log(y)


def roll(seq: ite, shift: int) -> ite:
    """
    Introduction
    ==========
    Roll all elements of a sequence

    Example
    ==========
    >>> roll([1, 2, 3, 4, 5, 6, 7, 8, 9], 5)
    [5, 6, 7, 8, 9, 1, 2, 3, 4]
    >>>
    :param seq: list | tuple | string
    :param shift: integer
    :return:
    """
    shift %= len(seq)
    return seq[-shift:] + seq[:-shift]


def root(x: num, y: num) -> num:
    """
    Introduction
    ==========
    Root \n
    ʸ√x̄

    Example
    ==========
    >>> root(2, 4)
    1.189207115002721
    >>>
    :param x: integer | float | complex
    :param y: integer | float | complex
    :return:
    """
    flag = False
    if x < 0 and int(y) == y and int(y) & 1:
        x = -x
        flag = True
    return -x ** (1 / y) if flag else x ** (1 / y)


def exp(x: real) -> real:
    """
    Introduction
    ==========
    Exponential \n
    eˣ

    Example
    ==========
    >>> exp(0.5)
    1.6487212707001282
    >>>
    :param x: integer | float
    :return:
    """
    return math.exp(x)


def ln(x: real) -> real:
    """
    Introduction
    ==========
    Natural logarithm \n
    logₑx

    Example
    ==========
    >>> ln(0.5)
    -0.6931471805599453
    >>>
    :param x: integer | float
    :return:
    """
    return math.log(x)


def gcd(*args: int) -> int:
    """
    Introduction
    ==========
    The greatest common factor of multiple integers

    Example
    ==========
    >>> gcd(12, 18, 24)
    6
    >>>
    :param args: integer
    :return:
    """
    if not args:
        return 0
    gcd_value = args[0]
    for n in args[1:]:
        gcd_value = math.gcd(gcd_value, n)
    return gcd_value


def lcm(*args: int) -> int:
    """
    Introduction
    ==========
    The least common multiple of multiple integers

    Example
    ==========
    >>> lcm(12, 18, 24)
    72
    >>>
    :param args: integer
    :return:
    """
    if not args:
        return 0
    try:
        f = math.lcm
    except AttributeError:
        def f(a, b):
            return a * b // math.gcd(a, b)
    lcm_value = args[0]
    for n in args[1:]:
        lcm_value = f(lcm_value, n)
    return lcm_value


def sin(x: real) -> real:
    """
    Sine
    :param x: integer | float
    :return:
    """
    return math.sin(x)


def cos(x: real) -> real:
    """
    Cosine
    :param x: integer | float
    :return:
    """
    return math.cos(x)


def tan(x: real) -> real:
    """
    Tangent
    :param x: integer | float
    :return:
    """
    return math.tan(x)


def csc(x: real) -> real:
    """
    Cosecant
    :param x: integer | float
    :return:
    """
    return 1 / sin(x)


def sec(x: real) -> real:
    """
    Secant
    :param x: integer | float
    :return:
    """
    return 1 / cos(x)


def cot(x: real) -> real:
    """
    Cotangent
    :param x: integer | float
    :return:
    """
    return cos(x) / sin(x)


def asin(x: real) -> real:
    """
    Arcsine
    :param x: integer | float
    :return:
    """
    return math.asin(x)


def acos(x: real) -> real:
    """
    Arccosine
    :param x: integer | float
    :return:
    """
    return math.acos(x)


def atan(x: real) -> real:
    """
    Arctangent
    :param x: integer | float
    :return:
    """
    return math.atan(x)


def acsc(x: real) -> real:
    """
    Arccosecant
    :param x: integer | float
    :return:
    """
    return asin(1 / x)


def asec(x: real) -> real:
    """
    Arcsecant
    :param x: integer | float
    :return:
    """
    return acos(1 / x)


def acot(x: real) -> real:
    """
    Arccotangent
    :param x: integer | float
    :return:
    """
    return atan(1 / x)


def sinh(x: real) -> real:
    """
    Hyperbolic Sine
    :param x: integer | float
    :return:
    """
    return math.sinh(x)


def cosh(x: real) -> real:
    """
    Hyperbolic Cosine
    :param x: integer | float
    :return:
    """
    return math.cosh(x)


def tanh(x: real) -> real:
    """
    Hyperbolic Tangent
    :param x: integer | float
    :return:
    """
    return math.tanh(x)


def csch(x: real) -> real:
    """
    Hyperbolic Cosecant
    :param x: integer | float
    :return:
    """
    return 1 / sinh(x)


def sech(x: real) -> real:
    """
    Hyperbolic Secant
    :param x: integer | float
    :return:
    """
    return 1 / cosh(x)


def coth(x: real) -> real:
    """
    Hyperbolic Cotangent
    :param x: integer | float
    :return:
    """
    return cosh(x) / sinh(x)


def asinh(x: real) -> real:
    """
    Hyperbolic Arcsine
    :param x: integer | float
    :return:
    """
    return math.asinh(x)


def acosh(x: real) -> real:
    """
    Hyperbolic Arccosine
    :param x: integer | float
    :return:
    """
    return math.acosh(x)


def atanh(x: real) -> real:
    """
    Hyperbolic Arctangent
    :param x: integer | float
    :return:
    """
    return math.atanh(x)


def acsch(x: real) -> real:
    """
    Hyperbolic Arccosecant
    :param x: integer | float
    :return:
    """
    return asinh(1 / x)


def asech(x: real) -> real:
    """
    Hyperbolic Arcsecant
    :param x: integer | float
    :return:
    """
    return acosh(1 / x)


def acoth(x: real) -> real:
    """
    Hyperbolic Arccotangent
    :param x: integer | float
    :return:
    """
    return atanh(1 / x)


def ptp(numbers: arr) -> num:
    """
    Introduction
    ==========
    Range of numbers

    Example
    ==========
    >>> ptp([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    9
    >>>
    :param numbers: list | tuple
    :return:
    """
    return max(numbers) - min(numbers)


def median(numbers: arr) -> num:
    """
    Introduction
    ==========
    Median of numbers

    Example
    ==========
    >>> median([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    5.5
    >>>
    :param numbers: list | tuple
    :return:
    """
    data = sorted(numbers)
    n = len(data)
    if n & 1:
        return data[n // 2]
    else:
        return (data[n // 2 - 1] + data[n // 2]) / 2


def freq(data: arr) -> dict:
    """
    Introduction
    ==========
    Frequency of data

    Example
    ==========
    >>> freq([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    {1: 1, 2: 1, 3: 1, 4: 1, 5: 1, 6: 1, 7: 1, 8: 1, 9: 1, 10: 1}
    >>>
    :param data: list | tuple
    :return:
    """
    result = {}
    for d in data:
        if d in result:
            result[d] += 1
        else:
            result[d] = 1
    return result


def mode(data: arr):
    """
    Introduction
    ==========
    Mode of data

    Example
    ==========
    >>> mode([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    1
    >>>
    :param data: list | tuple
    :return:
    """
    count_dict = freq(data)
    max_count = 0
    max_num = 0
    for k, v in count_dict.items():
        if v > max_count:
            max_count = v
            max_num = k
    return max_num


def mean(numbers: arr) -> num:
    """
    Introduction
    ==========
    Average of numbers

    Example
    ==========
    >>> mean([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    5.5
    >>>
    :param numbers: list | tuple
    :return:
    """
    return math.fsum(numbers) / len(numbers)


def geom_mean(numbers: arr) -> num:
    """
    Introduction
    ==========
    The geometric mean of numbers

    Example
    ==========
    >>> geom_mean([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    4.528728688116765
    >>>
    :param numbers: list | tuple
    :return:
    """
    return root(product(numbers), len(numbers))


def square_mean(numbers: arr) -> num:
    """
    Introduction
    ==========
    Square mean of numbers

    Example
    ==========
    >>> square_mean([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    6.2048368229954285
    >>>
    :param numbers: list | tuple
    :return:
    """
    return math.sqrt(math.fsum(map(lambda n: n * n, numbers)) / len(numbers))


def harm_mean(numbers: arr) -> num:
    """
    Introduction
    ==========
    The harmonic mean of numbers

    Example
    ==========
    >>> harm_mean([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    3.414171521474055
    >>>
    :param numbers: list | tuple
    :return:
    """
    return 0 if 0 in numbers else len(numbers) / math.fsum(map(lambda n: 1 / n, numbers))


def raw_moment(data: arr, order: int) -> float:
    """
    Introduction
    ==========
    Calculate the origin moment of the sample

    Example
    ==========
    >>> raw_moment([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 5)
    22082.5
    >>>
    :param data: list | tuple
    :param order: integer
    :return:
    """
    return math.fsum(map(lambda item: item ** order, data)) / len(data)


def central_moment(data: arr, order: int) -> float:
    """
    Introduction
    ==========
    Calculate the center moment of the sample

    Example
    ==========
    >>> central_moment([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 5)
    0.0
    >>>
    :param data: list | tuple
    :param order: integer
    :return:
    """
    avg = mean(data)
    return math.fsum(map(lambda item: (item - avg) ** order, data)) / len(data)


def var(numbers: arr, ddof: int = 0) -> num:
    """
    Introduction
    ==========
    Variance of numbers

    Example
    ==========
    >>> var([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    8.25
    >>>
    :param numbers: list | tuple
    :param ddof: integer
    :return:
    """
    if ddof >= len(numbers):
        return float("inf")
    avg = mean(numbers)

    def inner(item):
        d = item - avg
        return d * d

    return math.fsum(map(inner, numbers)) / (len(numbers) - ddof)


def skew(data: arr) -> float:
    """
    Introduction
    ==========
    Calculate the skewness of the sample

    Example
    ==========
    >>> skew([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    0.0
    >>>
    :param data: list | tuple
    :return:
    """
    return central_moment(data, 3) / std(data) ** 3


def kurt(data: arr, fisher: bool = True) -> float:
    """
    Introduction
    ==========
    Calculate the kurtosis of the sample

    Example
    ==========
    >>> kurt([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    -1.2242424242424244
    >>>
    :param data: list | tuple
    :param fisher: boolean
    :return:
    """
    result = central_moment(data, 4) / std(data) ** 4
    if fisher:
        result -= 3
    return result


def std(numbers: arr, ddof: int = 0) -> num:
    """
    Introduction
    ==========
    Standard deviation of numbers

    Example
    ==========
    >>> std([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    2.8722813232690143
    >>>
    :param numbers: list | tuple
    :param ddof: integer
    :return:
    """
    return math.sqrt(var(numbers, ddof=ddof))


def cov(x: arr, y: arr, ddof: int = 0) -> num:
    """
    Introduction
    ==========
    Covariance of numbers

    Example
    ==========
    >>> cov([1, 2, 3, 4, 5], [6, 7, 8, 9, 10])
    2.0
    >>>
    :param x: list | tuple
    :param y: list | tuple
    :param ddof: integer
    :return:
    """
    if ddof >= min(len(x), len(y)):
        return float("inf")
    mean_x = mean(x)
    mean_y = mean(y)
    return math.fsum([(x_i - mean_x) * (y_i - mean_y) for x_i, y_i in zip(x, y)]) / (len(x) - ddof)


def corr_coeff(x: arr, y: arr) -> num:
    """
    Introduction
    ==========
    The correlation coefficient of numbers

    Example
    ==========
    >>> corr_coeff([1, 2, 3, 4, 5], [6, 7, 8, 9, 10])
    0.9999999999999998
    >>>
    :param x: list | tuple
    :param y: list | tuple
    :return:
    """
    if len(set(y)) == 1:
        return 1.0
    n = len(x)
    mean_x = mean(x)
    mean_y = mean(y)
    covariance = math.fsum((_x - mean_x) * (_y - mean_y) for _x, _y in zip(x, y)) / n
    std_x = math.sqrt(math.fsum((_x - mean_x) ** 2 for _x in x) / n)
    std_y = math.sqrt(math.fsum((_y - mean_y) ** 2 for _y in y) / n)
    return covariance / (std_x * std_y)


def coeff_det(x: arr, y: arr) -> num:
    """
    Introduction
    ==========
    The coefficient of determination for numbers

    Example
    ==========
    >>> coeff_det([1, 2, 3, 4, 5], [6, 7, 8, 9, 10])
    0.9999999999999996
    >>>
    :param x: list | tuple
    :param y: list | tuple
    :return:
    """
    return corr_coeff(x, y) ** 2


def quantile(data: list, q: float, interpolation: str = "linear", ordered: bool = False) -> float:
    """
    Introduction
    ==========
    Compute the q-th quantile of the given data list.

    Example
    ==========
    >>> quantile([3, 6, 7, 8, 8, 10, 13, 15, 16, 20], 0.5)
    9.0
    >>>
    :param data: A list of numerical data.
    :param q: The quantile value to compute, which must be between 0 and 1 inclusive.
    :param interpolation: The interpolation method to use when the desired quantile lies between two data points.
        Valid options are 'linear', 'lower', 'higher', 'midpoint', and 'nearest'.
    :param ordered: A boolean flag indicating whether the input data is already sorted.
        If True, the data will not be sorted.
    :return: The computed q-th quantile of the data.
    """
    if not 0 <= q <= 1:
        raise ValueError("Quantile value q must be between 0 and 1 inclusive.")
    sorted_data = data if ordered else sorted(data)
    n = len(sorted_data)
    index = q * (n - 1)
    if index.is_integer():
        return sorted_data[int(index)]
    else:
        valid_interpolations = ["linear", "lower", "higher", "midpoint", "nearest"]
        if interpolation == "linear":
            lower = sorted_data[int(index)]
            upper = sorted_data[int(index) + 1]
            return lower + (upper - lower) * (index - int(index))
        if interpolation == "lower":
            return sorted_data[int(index)]
        if interpolation == "higher":
            return sorted_data[int(index) + 1]
        if interpolation == "midpoint":
            lower = sorted_data[int(index)]
            upper = sorted_data[int(index) + 1]
            return (lower + upper) / 2
        if interpolation == "nearest":
            lower_index = int(index)
            upper_index = lower_index + 1
            lower_diff = abs(index - lower_index)
            upper_diff = abs(upper_index - index)
            if lower_diff < upper_diff:
                return sorted_data[lower_index]
            else:
                return sorted_data[upper_index]
        raise ValueError("Invalid interpolation method. Choose from {}.".format(valid_interpolations))


def product(numbers: arr) -> num:
    """
    Introduction
    ==========
    Product of multiple numbers

    Example
    ==========
    >>> product([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    3628800
    >>>
    :param numbers: list | tuple
    :return:
    """
    if len(numbers) == 0:
        return 1
    elif 0 in numbers:
        return 0
    elif len(numbers) == 1:
        return numbers[0]
    elif len(numbers) <= 4096:
        result = numbers[0]
        for i in numbers[1:]:
            result *= i
        return result

    def prod(n):
        if len(n) == 1:
            return n[0]
        mid = len(n) // 2
        left_half = n[:mid]
        right_half = n[mid:]
        left_product = prod(left_half)
        right_product = prod(right_half)
        return left_product * right_product

    return prod(numbers)


def sigma(i: int, n: int, f) -> num:
    """
    Introduction
    ==========
    n \n
    Σ  f(x) \n
    i

    Example
    ==========
    >>> sigma(1, 10, lambda x: x ** x)
    10405071317
    >>>
    :param i: integer
    :param n: integer
    :param f: function
    :return:
    """
    return sum(map(f, range(i, n + 1)))


def pi(i: int, n: int, f) -> num:
    """
    Introduction
    ==========
    n \n
    Π  f(x) \n
    i

    Example
    ==========
    >>> pi(1, 10, lambda x: x ** x)
    215779412229418562091680268288000000000000000
    >>>
    :param i: integer
    :param n: integer
    :param f: function
    :return:
    """
    return product(list(map(f, range(i, n + 1))))


def deriv(f, x: float, h: float = 1e-6, method: str = "complex_step", *args, **kwargs):
    """
    Introduction
    ==========
    Compute the derivative of a function at a given point using various numerical methods.

    Example
    ==========
    >>> deriv(lambda a: a ** 2, 2)
    4.0
    >>>
    :param f: The function to differentiate.
    :param x: The point at which to evaluate the derivative.
    :param h: The step size for numerical differentiation.
    :param method: The numerical differentiation method to use.
        Supported methods include 'complex_step', 'central_difference', 'forward_difference', and 'backward_difference'.
    :return: The derivative of the function at the point x.
    """
    supported_methods = ["complex_step", "central_difference", "forward_difference", "backward_difference"]
    if method not in supported_methods:
        raise ValueError("Invalid method. Supported methods are: {}".format(supported_methods))
    if method == "complex_step":
        return f(x + 1j * h, *args, **kwargs).imag / h
    elif method == "central_difference":
        h /= 2
        return (f(x + h, *args, **kwargs) - f(x - h, *args, **kwargs)) / (2 * h)
    elif method == "forward_difference":
        return (f(x + h, *args, **kwargs) - f(x, *args, **kwargs)) / h
    elif method == "backward_difference":
        return (f(x, *args, **kwargs) - f(x - h, *args, **kwargs)) / h


def integ(f, x_start: real, x_end: real, n: int = 1000000, *args, **kwargs) -> float:
    """
    Introduction
    ==========
    Compute the integral of a function over a specified interval using the composite Simpson's rule.

    Example
    ==========
    >>> integ(lambda a: a ** 2, 0, 2)
    2.6666666666666665
    >>>
    :param f: The function to integrate.
    :param x_start: The starting point of the interval.
    :param x_end: The ending point of the interval.
    :param n: The number of subintervals to use.
    :return: The numerical integral of the function over the interval.
    """
    h = (x_end - x_start) / n
    if args or kwargs:
        s = f(x_start, *args, **kwargs) + f(x_end, *args, **kwargs)
        s += math.fsum(map(lambda i: 4 * f(x_start + i * h, *args, **kwargs), range(1, n, 2)))
        s += math.fsum(map(lambda i: 2 * f(x_start + i * h, *args, **kwargs), range(2, n - 1, 2)))
    else:
        s = f(x_start) + f(x_end)
        s += math.fsum(map(lambda i: 4 * f(x_start + i * h), range(1, n, 2)))
        s += math.fsum(map(lambda i: 2 * f(x_start + i * h), range(2, n - 1, 2)))
    return s * h / 3


def beta(p: real, q: real) -> real:
    """
    Introduction
    ==========
    Beta function calculation \n
    Β(p, q)

    Example
    ==========
    >>> beta(0.5, 0.5)
    3.1415926535897927
    >>>
    :param p: integer | float
    :param q: integer | float
    :return:
    """
    return gamma(p) * gamma(q) / gamma(p + q)


def gamma(alpha: real) -> float:
    """
    Introduction
    ==========
    Gamma function calculation \n
    Γ(s)

    Example
    ==========
    >>> gamma(1.5)
    0.886226925452758
    >>>
    :param alpha: integer | float
    :return:
    """
    return math.gamma(alpha)


def factorial(n: int) -> int:
    """
    Introduction
    ==========
    Integer factorial

    Example
    ==========
    >>> factorial(10)
    3628800
    >>>
    :param n: integer
    :return:
    """
    return math.factorial(n)


def arrangement(n: int, r: int) -> int:
    """
    Introduction
    ==========
    A(n, r)

    Example
    ==========
    >>> arrangement(10, 5)
    30240
    >>>
    :param n: integer
    :param r: integer
    :return:
    """
    return product(list(range(n, n - r, -1)))


def combination(n: int, r: int) -> int:
    """
    Introduction
    ==========
    C(n, r)

    Example
    ==========
    >>> combination(10, 5)
    252
    >>>
    :param n: integer
    :param r: integer
    :return:
    """
    if r > n >> 1:
        return combination(n, n - r)
    return product(list(range(n, n - r, -1))) // product(list(range(r, 0, -1)))


def zeta(alpha: num) -> num:
    """
    Introduction
    ==========
    Zeta function calculation \n
    ζ(s)

    Example
    ==========
    >>> zeta(3)
    1.2020569031590942
    >>>
    :param alpha: integer | float | complex
    :return:
    """
    seq = map(lambda _: _ ** (-alpha), range(1, 1000000))
    if alpha.imag:
        return sum(seq)
    else:
        return math.fsum(seq)


def erf(x: real) -> float:
    """
    Introduction
    ==========
    Error function calculation

    Example
    ==========
    >>> erf(1)
    0.842700792949715
    >>>
    :param x: integer | float
    :return:
    """
    return math.erf(x)


def sigmoid(x: real) -> float:
    """
    Introduction
    ==========
    Sigmoid function calculation

    Example
    ==========
    >>> sigmoid(1)
    0.7310585786300049
    >>>
    :param x: integer | float
    :return:
    """
    return 1 / (1 + math.exp(-x))


def sign(x: num) -> num:
    """
    Introduction
    ==========
    Sign function calculation

    Example
    ==========
    >>> sign(10)
    1.0
    >>>
    :param x: integer | float | complex
    :return:
    """
    return 0.0 if x == 0 else x / abs(x)


def parity(x: int) -> int:
    """
    Introduction
    ==========
    Calculate (-1) ** n using bits

    Example
    ==========
    >>> parity(10)
    1
    >>>
    :param x: integer
    :return:
    """
    return (~x & 1) * 2 - 1


def cumsum(lst: arr) -> list:
    """
    Introduction
    ==========
    Sequence accumulation

    Example
    ==========
    >>> cumsum([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    [1, 3, 6, 10, 15, 21, 28, 36, 45, 55]
    >>>
    :param lst: list | tuple
    :return:
    """
    if not lst:
        return []
    cumulative_sum = [lst[0]]
    length = len(lst)
    for i in range(1, length):
        cumulative_sum.append(cumulative_sum[i - 1] + lst[i])
    return cumulative_sum


def cumprod(lst: arr) -> list:
    """
    Introduction
    ==========
    Sequence multiplication

    Example
    ==========
    >>> cumprod([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    [1, 2, 6, 24, 120, 720, 5040, 40320, 362880, 3628800]
    >>>
    :param lst: list | tuple
    :return:
    """
    if not lst:
        return []
    cumulative_prod = [lst[0]]
    length = len(lst)
    for i in range(1, length):
        cumulative_prod.append(cumulative_prod[i - 1] * lst[i])
    return cumulative_prod


def iroot(y: int, n: int) -> int:
    """
    Introduction
    ==========
    Round after rooting \n
    [ʸ√x̄]

    Example
    ==========
    >>> iroot(-2 ** 1000, 9)
    -2803995394640233757514136098827399
    >>>
    :param y: integer
    :param n: integer
    :return:
    """
    i = False
    if y < 0 and n & 1:
        y = abs(y)
        i = True
    if y in [0, 1] or n == 1:
        return y
    elif n >= y.bit_length():
        return 1
    try:
        g = int(y ** (1 / n) + 0.5)
    except OverflowError:
        e = math.log(y, 2) / n
        if e > 53:
            s = int(e - 53)
            g = int(2 ** (e - s) + 1) << s
        else:
            g = int(2 ** e)
    if g > 2 ** 50:
        p, x = -1, g
        while abs(x - p) >= 2:
            p, x = x, ((n - 1) * x + y // x ** (n - 1)) // n
    else:
        x = g
    t = x ** n
    while t < y:
        x += 1
        t = x ** n
    if t > y:
        x -= 1
    return -x if i else x


def totient(n: int) -> int:
    """
    Introduction
    ==========
    Euler's totient function

    Example
    ==========
    >>> totient(100)
    40
    >>>
    :param n: integer
    :return:
    """
    if n == 1:
        return 1
    phi = n
    for p in range(2, int(math.sqrt(n)) + 1):
        if n % p == 0:
            phi = phi - phi // p
            while n % p == 0:
                n //= p
    if n > 1:
        phi = phi - phi // n
    return phi


def mod_order(a: int, n: int, b: int) -> int:
    """
    Introduction
    ==========
    The order of element b in the multiplication group of module a

    Example
    ==========
    >>> mod_order(101, 100, 99)
    100
    >>>
    :param a: integer
    :param n: integer
    :param b: integer
    :return:
    """
    p = 1
    c = b % a
    while p <= n and c != 1:
        c = c * b % a
        p += 1
    if p <= n:
        return p
    else:
        return -1


def primitive_root(a: int, single: bool = False) -> Union[int, list]:
    """
    Introduction
    ==========
    The order of element b in the multiplication group of module a

    Example
    ==========
    >>> primitive_root(11)
    [2, 6, 7, 8]
    >>>
    :param a: integer
    :param single: bool
    :return:
    """
    n = totient(a)
    p = []
    for b in range(2, a):
        if mod_order(a, n, b) == n:
            if single:
                return b
            p.append(b)
    return p


def normalize(data: arr, target: num = 1) -> arr:
    """
    Introduction
    ==========
    Scale all numbers proportionally until the total is exactly the target value

    Example
    ==========
    >>> normalize([1, 2, 3, 4, 5])
    [0.06666666666666667, 0.13333333333333333, 0.2, 0.26666666666666666, 0.3333333333333333]
    >>>
    :param data: list | tuple
    :param target: integer | float
    :return:
    """
    total = math.fsum(data)
    if total == 0 or target == 0:
        return type(data)([0]) * len(data)
    ratio = target / total
    return type(data)(map(lambda x: x * ratio, data))


def average(data: arr, weights: arr) -> float:
    """
    Introduction
    ==========
    Calculate the weighted average or expected value of numbers

    Example
    ==========
    >>> average([1, 2, 3, 4, 5], [6, 7, 8, 9, 10])
    3.25
    >>>
    :param data: list | tuple
    :param weights: list | tuple
    :return:
    """
    result = math.fsum(map(lambda value: value[0] * value[1], zip(data, weights)))
    return result / math.fsum(weights)


def exgcd(a: int, b: int) -> tuple:
    """
    Introduction
    ==========
    Extended Euclidean algorithm for calculating the maximum common factor and Bezier number

    Example
    ==========
    >>> exgcd(142, 857)
    (1, -344, 57)
    >>>
    :param a: integer
    :param b: integer
    :return:
    """
    if b == 0:
        return a, 1, 0
    else:
        x, y, g = 0, 1, b
        x1, y1, gcd_prev = 1, 0, a
        while g != 0:
            q = gcd_prev // g
            gcd_next, x_next, y_next = gcd_prev % g, x1 - q * x, y1 - q * y
            gcd_prev, x1, y1 = g, x, y
            g, x, y = gcd_next, x_next, y_next
        return gcd_prev, x1, y1


def crt(n: arr, a: arr) -> int:
    """
    Introduction
    ==========
    Using Chinese residual theory to solve congruence equations

    Example
    ==========
    >>> crt([9, 8, 7], [6, 5, 4])
    501
    >>>
    :param n: list | tuple
    :param a: list | tuple
    :return:
    """
    s = 0
    prod = product(n)
    for ni, ai in zip(n, a):
        if ni <= ai:
            return 0
        p = prod // ni
        g, x, y = exgcd(p, ni)
        if g != 1:
            return 0
        else:
            inv = x % ni
        s += ai * inv * p
    return s % prod


def isqrt(x: int) -> int:
    """
    Introduction
    ==========
    Round after square root \n
    [√x̄]

    Example
    ==========
    >>> isqrt(620448401733239439360000)
    787685471322
    >>>
    :param x: integer
    :return:
    """
    return iroot(x, 2)


def is_possibly_square(n: int) -> bool:
    """
    Introduction
    ==========
    Check if it is a possible square number

    Example
    ==========
    >>> is_possibly_square(123456 ** 789)
    True
    >>>
    :param n: integer
    :return:
    """
    if n < 0:
        return False
    if n % 720 not in __mod:
        return False
    if n & 0xFF not in __and:
        return False
    return True


def is_square(n: int) -> bool:
    """
    Introduction
    ==========
    Check if it is a square number

    Example
    ==========
    >>> is_square(123456 ** 789)
    False
    >>>
    :param n: integer
    :return:
    """
    if is_possibly_square(n):
        try:
            from math import isqrt as f
        except ImportError:
            f = isqrt
        sqrt = f(n)
        return sqrt * sqrt == n
    return False


A = arrangement
C = combination
if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=True)



class BaseGraph:
    def __init__(self):
        self.graph = {}

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.to_list())

    def to_list(self):
        if isinstance(self, BaseWeGraph):
            return sorted(map(lambda item: (item[0], sorted(item[1].items())), self.graph.items()))
        else:
            return sorted(self.graph.items())

    def add_vertex(self, vertex):
        if vertex not in self.graph:
            self.graph[vertex] = []

    def remove_vertex(self, vertex):
        if vertex in self.graph:
            neighbors = self.graph[vertex]
            del self.graph[vertex]
            for neighbor in neighbors:
                self.graph[neighbor].remove(vertex)

    def has_vertex(self, vertex):
        return vertex in self.graph

    def has_edge(self, vertex1, vertex2):
        return vertex1 in self.graph and vertex2 in self.graph[vertex1]

    def get_edges(self, vertex):
        return self.graph.get(vertex, {})

    def all_vertices(self):
        return sorted(set(list(self.graph) + sum([list(self.graph[vertex]) for vertex in self.graph], [])))

    def all_edges(self):
        edges = []
        for vertex, neighbours in self.graph.items():
            for neighbour in neighbours:
                edges.append((vertex, neighbour))
        return sorted(edges)

    def to_adjacency_matrix(self):
        vertices = self.all_vertices()
        num_vertices = len(vertices)
        matrix = [[0] * num_vertices for _ in range(num_vertices)]
        if isinstance(self, BaseWeGraph):
            for u in self.graph:
                for v in self.graph[u].items():
                    try:
                        v, w = v
                        matrix_index_u = vertices.index(u)
                        matrix_index_v = vertices.index(v)
                        matrix[matrix_index_u][matrix_index_v] = w
                    except ValueError:
                        continue
        else:
            for u in self.graph:
                for v in self.graph[u]:
                    try:
                        matrix_index_u = vertices.index(u)
                        matrix_index_v = vertices.index(v)
                        matrix[matrix_index_u][matrix_index_v] = 1
                    except ValueError:
                        continue
        return [[self.__class__.__name__] + vertices] + [[vertex] + row for vertex, row in zip(vertices, matrix)]

    def dfs(self, start_vertex, visited=None):
        if visited is None:
            visited = set()
        visited.add(start_vertex)
        search_list = [start_vertex]
        if start_vertex not in self.graph:
            return [start_vertex]
        for neighbour in self.graph[start_vertex]:
            if neighbour not in visited:
                search_list.extend(self.dfs(neighbour, visited))
        return search_list

    def bfs(self, start_vertex):
        visited = {start_vertex}
        queue = [start_vertex]
        search_list = [start_vertex]
        while queue:
            vertex = queue.pop(0)
            if vertex in self.graph:
                for neighbour in self.graph[vertex]:
                    if neighbour not in visited:
                        visited.add(neighbour)
                        queue.append(neighbour)
                        search_list.append(neighbour)
        return search_list

    def is_connected(self):
        visited = set()
        queue = []
        for node in self.graph:
            if node not in visited:
                visited.add(node)
                queue.append(node)
                break
        while queue:
            node = queue.pop(0)
            for neighbor in self.graph[node]:
                if neighbor not in visited:
                    visited.add(neighbor)
                    queue.append(neighbor)
        return len(visited) == len(self.graph)

    def is_complete(self):
        num_nodes = len(self.graph)
        return all([len(self.graph[node]) == num_nodes - 1 for node in self.graph])


class DiGraph(BaseGraph):
    def add_edge(self, vertex1, vertex2):
        if vertex1 in self.graph:
            self.graph[vertex1].append(vertex2)
        else:
            self.graph[vertex1] = [vertex2]

    def remove_edge(self, vertex1, vertex2):
        if vertex1 in self.graph and vertex2 in self.graph[vertex1]:
            self.graph[vertex1].remove(vertex2)
            if not self.graph[vertex1]:
                del self.graph[vertex1]

    def __str__(self):
        edges = []
        for vertex, neighbours in self.graph.items():
            for neighbour in neighbours:
                edges.append("{} -> {}".format(vertex, neighbour))
        return "\n".join(sorted(edges))


class UnGraph(BaseGraph):
    def add_edge(self, vertex1, vertex2):
        if vertex1 in self.graph:
            self.graph[vertex1].append(vertex2)
        else:
            self.graph[vertex1] = [vertex2]
        if vertex2 in self.graph:
            self.graph[vertex2].append(vertex1)
        else:
            self.graph[vertex2] = [vertex1]

    def remove_edge(self, vertex1, vertex2):
        if vertex1 in self.graph and vertex2 in self.graph:
            if vertex2 in self.graph[vertex1]:
                self.graph[vertex1].remove(vertex2)
            if vertex1 in self.graph[vertex2]:
                self.graph[vertex2].remove(vertex1)
            if not self.graph[vertex1]:
                del self.graph[vertex1]
            if not self.graph[vertex2]:
                del self.graph[vertex2]

    def __str__(self):
        edges = []
        for vertex, neighbours in self.graph.items():
            for neighbour in neighbours:
                edges.append("{} -- {}".format(vertex, neighbour))
        return "\n".join(sorted(edges))


class BaseWeGraph(BaseGraph):
    def add_edge(self, vertex1, vertex2, weight=0):
        raise NotImplementedError

    def remove_edge(self, vertex1, vertex2):
        raise NotImplementedError

    def add_vertex(self, vertex):
        if vertex not in self.graph:
            self.graph[vertex] = {}

    def remove_vertex(self, vertex):
        if vertex in self.graph:
            neighbors = self.graph[vertex]
            del self.graph[vertex]
            for neighbor in neighbors:
                self.remove_edge(neighbor, vertex)

    def get_edge_weight(self, vertex1, vertex2):
        if self.has_edge(vertex1, vertex2):
            return self.graph[vertex1][vertex2]

    def get_in_degree_weight_sum(self, vertex):
        return sum([self.get_edge_weight(neighbor, vertex)
                    for neighbor in self.graph if vertex in self.graph[neighbor]])

    def get_out_degree_weight_sum(self, vertex):
        return sum([self.get_edge_weight(vertex, neighbor) for neighbor in self.graph.get(vertex, {})])

    def dijkstra(self, start):
        distances = {vertex: float("inf") for vertex in self.graph}
        distances[start] = 0
        predecessors = {vertex: None for vertex in self.graph}
        queue = [(0, start)]
        visited = set()
        while queue:
            current_distance, current_vertex = min(queue, key=lambda x: x[0])
            if current_vertex in visited:
                queue.remove((current_distance, current_vertex))
                continue
            visited.add(current_vertex)
            queue.remove((current_distance, current_vertex))
            for neighbor, weight in self.graph[current_vertex].items():
                distance = current_distance + weight
                if distance < distances[neighbor]:
                    distances[neighbor] = distance
                    predecessors[neighbor] = current_vertex
                    queue.append((distance, neighbor))
        return distances, predecessors

    def reconstruct_path(self, start, end):
        distances, predecessors = self.dijkstra(start)
        path = []
        current = end
        while current is not None:
            path.append(current)
            current = predecessors.get(current)
        return path[::-1] if path else None


class WeDiGraph(BaseWeGraph):
    def add_edge(self, vertex1, vertex2, weight=0):
        if vertex1 not in self.graph:
            self.graph[vertex1] = {}
        self.graph[vertex1][vertex2] = weight

    def remove_edge(self, vertex1, vertex2):
        if vertex1 in self.graph and vertex2 in self.graph[vertex1]:
            del self.graph[vertex1][vertex2]
            if not self.graph[vertex1]:
                del self.graph[vertex1]

    def __str__(self):
        edges = []
        for vertex, neighbours in self.graph.items():
            for neighbour, weight in neighbours.items():
                edges.append("{} -[{}]> {}".format(vertex, weight, neighbour))
        return "\n".join(sorted(edges))


class WeUnGraph(BaseWeGraph):
    def add_edge(self, vertex1, vertex2, weight=0):
        if vertex1 not in self.graph:
            self.graph[vertex1] = {}
        if vertex2 not in self.graph:
            self.graph[vertex2] = {}
        self.graph[vertex1][vertex2] = weight
        self.graph[vertex2][vertex1] = weight

    def remove_edge(self, vertex1, vertex2):
        if vertex1 in self.graph and vertex2 in self.graph[vertex1]:
            del self.graph[vertex1][vertex2]
            del self.graph[vertex2][vertex1]
            if not self.graph[vertex1]:
                del self.graph[vertex1]
            if vertex1 != vertex2 and not self.graph[vertex2]:
                del self.graph[vertex2]

    def __str__(self):
        edges = []
        for vertex, neighbours in self.graph.items():
            for neighbour, weight in neighbours.items():
                edges.append("{} -[{}]- {}".format(vertex, weight, neighbour))
        return "\n".join(sorted(edges))
